/*
    Neighand: Neighborhood Handling library

    The goal of this class is to find 3D neighbors efficiently.
    It stores the position of objects in 3D. The user can then ask what
    are the neighbor objects for any given point, within a radius
    (aka locality query, or neighborhood search).

    Read the documentation contained in the article "Spherical
    Indexing for Neighborhood Queries" for full details.

    This file defines the body of the member functions for the main class.
    It is either included directly from the neighand.h file,
    or from the cpp in case of library specialization.

    Nicolas Brodu, 2006
    Code released according to the GNU GPL, v2 or above.
*/

#include <string.h>
#include <math.h>

#ifdef NEIGHAND_SPECIALIZE
// This is an independent CPP when specializing, it includes the main header
#include "neighand.h"
#define NEIGHAND_TEMPLATE_ARGUMENTS NEIGHAND_SPECIALIZE
#else
// And this is an included header file for the generic version
#define NEIGHAND_TEMPLATE_ARGUMENTS exp2divx,exp2divy,exp2divz,wrapX,wrapY,wrapZ,layerZ
#endif

namespace neighand {

#ifdef NEIGHAND_SPECIALIZE
template<> NeighborhoodHandler<NEIGHAND_SPECIALIZE>::NeighborhoodHandler(FloatType minx, FloatType miny, FloatType minz, FloatType cellSize, FloatType queryRevertSizeCellUnit, FloatType closestRevertSizeCellUnit, FloatType sphereSelectionBias, const char* precompFile)
#else
template <int exp2divx, int exp2divy, int exp2divz, bool wrapX, bool wrapY, bool wrapZ, bool layerZ>
NeighborhoodHandler<NEIGHAND_TEMPLATE_ARGUMENTS>::NeighborhoodHandler(FloatType minx, FloatType miny, FloatType minz, FloatType cellSize, FloatType queryRevertSizeCellUnit, FloatType closestRevertSizeCellUnit, FloatType sphereSelectionBias, const char* precompFile)
#endif
: helper(minx, miny, minz, cellSize)
{

    queryRevertSize = queryRevertSizeCellUnit * cellSize;
    closestRevertSize = closestRevertSizeCellUnit * cellSize;

#ifdef NEIGHAND_SELECT_METHOD_STAT
    resetStat();
#endif

    cellSizeSqInv = FloatType(1.0f) / (cellSize*cellSize);

    // The sphere offsets pre-computation depends on wrapping
    WrapHelper<NEIGHAND_TEMPLATE_ARGUMENTS>::buildDistanceArray(baseDistanceArray, sphereOffsets, shavingComplement, shavingOffsets, maxWorldDist32, distanceVolumeTableBase, precompFile);

    distanceVolumeTable = new uint32_t[maxWorldDist32+1];
    setSphereSelectionBias(sphereSelectionBias);

    // Now, create the object structure
    void* newMem;
    posix_memalign(&newMem, 64, WrapHelper<NEIGHAND_TEMPLATE_ARGUMENTS>::ArraySize*sizeof(CellEntry));
    cells = reinterpret_cast<CellEntry*>(newMem);
    // 64-bit alignment is not needed here, but cannot hurt
    posix_memalign(&newMem, 64, WrapHelper<NEIGHAND_TEMPLATE_ARGUMENTS>::ArraySize*sizeof(ObjectProxy*));
    cachedLists = reinterpret_cast<ObjectProxy**>(newMem);

    // prepare array
    for (uint_fast32_t i=0; i<WrapHelper<NEIGHAND_TEMPLATE_ARGUMENTS>::ArraySize; ++i) {
        // start empty
        cells[i].objects = 0;
        cells[i].nextNonEmpty = 0;
        // cache is empty too
        cachedLists[i] = 0;
    }
    // Done, array is empty and ready to serve

    // Cache the real outside cell reference, no need to get from array index each time
    // The index at 2^(exp2divx+exp2divy+exp2divz) is the dummy cell, it is not a valid main region index
    // The index just after is the real outside cell
    // => the dummy index is used to avoid duplicates when the query sphere intersects the outside region
    // => the dummy index is the "virtual" cell that would be at that position of the sphere outside, but all are mapped to the empty one.
    // This function has no effect when there is no outside cell, so in that case it doesn't matter that the indice is wrong
    helper.setOutsideCell(&cells[WrapHelper<NEIGHAND_TEMPLATE_ARGUMENTS>::ArraySize-1]);

    firstNonEmptyCell = 0; totalNonEmpty = 0;
}


// Hopefully C++ 0x will have support for chaining constructors!
#ifdef NEIGHAND_SPECIALIZE
template<> NeighborhoodHandler<NEIGHAND_SPECIALIZE>::NeighborhoodHandler(FloatType minx, FloatType miny, FloatType minz, FloatType cellSize, const char* precompFile)
#else
template <int exp2divx, int exp2divy, int exp2divz, bool wrapX, bool wrapY, bool wrapZ, bool layerZ>
NeighborhoodHandler<NEIGHAND_TEMPLATE_ARGUMENTS>::NeighborhoodHandler(FloatType minx, FloatType miny, FloatType minz, FloatType cellSize, const char* precompFile)
#endif
: helper(minx, miny, minz, cellSize)
{
    // See other constructor for comments. Duplicated code, booo-ooo, that's bad!!!
    queryRevertSize = 0.0f;
    closestRevertSize = 0.0f;
#ifdef NEIGHAND_SELECT_METHOD_STAT
    resetStat();
#endif
    cellSizeSqInv = FloatType(1.0f) / (cellSize*cellSize);
    WrapHelper<NEIGHAND_TEMPLATE_ARGUMENTS>::buildDistanceArray(baseDistanceArray, sphereOffsets, shavingComplement, shavingOffsets, maxWorldDist32, distanceVolumeTableBase, precompFile);
    distanceVolumeTable = new uint32_t[maxWorldDist32+1];
    setSphereSelectionBias(1.0f);
    void* newMem;
    posix_memalign(&newMem, 64, WrapHelper<NEIGHAND_TEMPLATE_ARGUMENTS>::ArraySize*sizeof(CellEntry));
    cells = reinterpret_cast<CellEntry*>(newMem);
    posix_memalign(&newMem, 64, WrapHelper<NEIGHAND_TEMPLATE_ARGUMENTS>::ArraySize*sizeof(ObjectProxy*));
    cachedLists = reinterpret_cast<ObjectProxy**>(newMem);
    for (uint_fast32_t i=0; i<WrapHelper<NEIGHAND_TEMPLATE_ARGUMENTS>::ArraySize; ++i) {
        cells[i].objects = 0;
        cells[i].nextNonEmpty = 0;
        cachedLists[i] = 0;
    }
    helper.setOutsideCell(&cells[WrapHelper<NEIGHAND_TEMPLATE_ARGUMENTS>::ArraySize-1]);
    firstNonEmptyCell = 0; totalNonEmpty = 0;
}



#ifdef NEIGHAND_SPECIALIZE
template<> NeighborhoodHandler<NEIGHAND_SPECIALIZE>::~NeighborhoodHandler()
#else
template <int exp2divx, int exp2divy, int exp2divz, bool wrapX, bool wrapY, bool wrapZ, bool layerZ>
NEIGHAND_INLINE NeighborhoodHandler<NEIGHAND_TEMPLATE_ARGUMENTS>::~NeighborhoodHandler()
#endif
{
    // cleanup arrays allocated with posix_memalign
    free(cachedLists);
    free(cells);
    free(baseDistanceArray-1);
    free(sphereOffsets);
    for (int shaving = 0; shaving < 64; ++shaving) {
        free(shavingComplement[shaving]-1);
        free(shavingOffsets[shaving]);
    }
    delete [] distanceVolumeTable;
    delete [] distanceVolumeTableBase;
}

#ifdef NEIGHAND_SPECIALIZE
template<> void NeighborhoodHandler<NEIGHAND_SPECIALIZE>::insertNoAllocate(FloatType x, FloatType y, FloatType z, ObjectProxy* proxy, uint32_t idx)
#else
template <int exp2divx, int exp2divy, int exp2divz, bool wrapX, bool wrapY, bool wrapZ, bool layerZ>
NEIGHAND_INLINE void NeighborhoodHandler<NEIGHAND_TEMPLATE_ARGUMENTS>::insertNoAllocate(FloatType x, FloatType y, FloatType z, ObjectProxy* proxy, uint32_t idx)
#endif
{
    CellEntry* cell = &cells[idx];

    *reinterpret_cast<uint32_t*>(&proxy->x) = *reinterpret_cast<uint32_t*>(&x);
    *reinterpret_cast<uint32_t*>(&proxy->y) = *reinterpret_cast<uint32_t*>(&y);
    *reinterpret_cast<uint32_t*>(&proxy->z) = *reinterpret_cast<uint32_t*>(&z);

    proxy->next = cell->objects;
    proxy->prev = 0;
    cell->objects = proxy;
    if (proxy->next) proxy->next->prev = proxy;

    // Was this an empty cell ? => Maintain non-empty list if needed
    if ((!cachedLists[idx]) && WrapHelper<NEIGHAND_TEMPLATE_ARGUMENTS>::isInside(idx)) {
        if ((firstNonEmptyCell==0) || (cell < firstNonEmptyCell)) {
            cell->nextNonEmpty = firstNonEmptyCell;
            firstNonEmptyCell = cell;
        } else {
            // prev will find first object in worse case, hence cannot go <0
            CellEntry* prev = cell-1; while (prev->objects==0) --prev;
            cell->nextNonEmpty = prev->nextNonEmpty;
            prev->nextNonEmpty = cell;
        }
        ++totalNonEmpty;
    }

    // update cache
    cachedLists[idx] = cell->objects;

    // update proxy info, increase counts
    proxy->cellIndex = idx;
    proxy->cell = cell;
}


#ifdef NEIGHAND_SPECIALIZE
template<> ObjectProxy* NeighborhoodHandler<NEIGHAND_SPECIALIZE>::insert(FloatType x, FloatType y, FloatType z, void* object)
#else
template <int exp2divx, int exp2divy, int exp2divz, bool wrapX, bool wrapY, bool wrapZ, bool layerZ>
NEIGHAND_INLINE ObjectProxy* NeighborhoodHandler<NEIGHAND_TEMPLATE_ARGUMENTS>::insert(FloatType x, FloatType y, FloatType z, void* object)
#endif
{
    // Find cell for that x/y/z position
    uint32_t idx = helper.getCellIndexForWorldPosition(x, y, z);
    // get a free proxy.
    ObjectProxy* proxy = lineManager.getProxy();
    proxy->object = object;
    // call internal routine
    insertNoAllocate(x, y, z, proxy, idx);
    return proxy;
}

#ifdef NEIGHAND_SPECIALIZE
template<> void NeighborhoodHandler<NEIGHAND_SPECIALIZE>::removeNoDealloc(ObjectProxy* proxy)
#else
template <int exp2divx, int exp2divy, int exp2divz, bool wrapX, bool wrapY, bool wrapZ, bool layerZ>
NEIGHAND_INLINE void NeighborhoodHandler<NEIGHAND_TEMPLATE_ARGUMENTS>::removeNoDealloc(ObjectProxy* proxy)
#endif
{
    CellEntry* cell = proxy->cell;

    if (proxy->prev) proxy->prev->next = proxy->next;
    if (proxy->next) proxy->next->prev = proxy->prev;

    if (cell->objects == proxy) {
        cell->objects = proxy->next;
        cachedLists[proxy->cellIndex] = proxy->next;
        if (proxy->next) proxy->next->prev = 0;

        // Is this now an empty cell ? => Maintain non-empty list if needed
        if ((!proxy->next) && WrapHelper<NEIGHAND_TEMPLATE_ARGUMENTS>::isInside(proxy->cellIndex)) {
            if (totalNonEmpty>1) {
                if (proxy->cell <= firstNonEmptyCell) {
                    firstNonEmptyCell = cell->nextNonEmpty;
                } else {
                    // prev will find first object in worse case, hence cannot go <0
                    CellEntry* prev = cell-1; while (prev->objects==0) --prev;
                    prev->nextNonEmpty = cell->nextNonEmpty;
                }
            } else firstNonEmptyCell = 0;
            --totalNonEmpty;
        }
    }

}

#ifdef NEIGHAND_SPECIALIZE
template<> void NeighborhoodHandler<NEIGHAND_SPECIALIZE>::remove(ObjectProxy* proxy)
#else
template <int exp2divx, int exp2divy, int exp2divz, bool wrapX, bool wrapY, bool wrapZ, bool layerZ>
NEIGHAND_INLINE void NeighborhoodHandler<NEIGHAND_TEMPLATE_ARGUMENTS>::remove(ObjectProxy* proxy)
#endif
{
    // internal remove, common with update
    removeNoDealloc(proxy);
    // and release the proxy for external API
    lineManager.releaseProxy(proxy);
}


#ifdef NEIGHAND_SPECIALIZE
template<> void NeighborhoodHandler<NEIGHAND_SPECIALIZE>::update(FloatType x, FloatType y, FloatType z, ObjectProxy* proxy)
#else
template <int exp2divx, int exp2divy, int exp2divz, bool wrapX, bool wrapY, bool wrapZ, bool layerZ>
NEIGHAND_INLINE void NeighborhoodHandler<NEIGHAND_TEMPLATE_ARGUMENTS>::update(FloatType x, FloatType y, FloatType z, ObjectProxy* proxy)
#endif
{

    uint32_t idx = helper.getCellIndexForWorldPosition(x, y, z);

    // if this is still the same cell index, update x/y/z and done!
    if (idx==proxy->cellIndex) {
        // float affectations are slower => byte copy
        *reinterpret_cast<uint32_t*>(&proxy->x) = *reinterpret_cast<uint32_t*>(&x);
        *reinterpret_cast<uint32_t*>(&proxy->y) = *reinterpret_cast<uint32_t*>(&y);
        *reinterpret_cast<uint32_t*>(&proxy->z) = *reinterpret_cast<uint32_t*>(&z);
        return;
    }

    // Otherwise, need to change cell.

    // first remove from current cell
    removeNoDealloc(proxy);

    // Now insert in new cell
    insertNoAllocate(x, y, z, proxy, idx);
}

// Main query routine: apply functor to all neighbors
// Use separate logical file for maintenance
#include "neighand_apply.hpp"

// Again, using x/y/z API
#define NEIGHAND_APPLY_XYZ_API 1
#include "neighand_apply.hpp"
#undef NEIGHAND_APPLY_XYZ_API

// Again, using C callback API
#define NEIGHAND_C_CALLBACK_API 1
#include "neighand_apply.hpp"

#define NEIGHAND_APPLY_XYZ_API 1
#include "neighand_apply.hpp"
#undef NEIGHAND_APPLY_XYZ_API

#undef NEIGHAND_C_CALLBACK_API


#include "neighand_apply_all.hpp"

#define NEIGHAND_C_CALLBACK_API 1
#include "neighand_apply_all.hpp"
#undef NEIGHAND_C_CALLBACK_API


namespace detail {

// Functor to build neighbor lists, without an object to avoid
struct ListBuilderNoAvoid {
    NeighborListHolder* nlh;
    NeighborList* endCell;
    NEIGHAND_INLINE ListBuilderNoAvoid(NeighborListHolder* _nlh) : nlh(_nlh), endCell(0) {}

    NEIGHAND_INLINE void operator()(void* object) NEIGHAND_ALWAYS_INLINE {
        // need to allocate new line?
        // 8 cells per line => check alignement. Also works for initial case with 0
        if ((nlh->size&7)==0) {
            if (endCell) {
                endCell->next = reinterpret_cast<NeighborList*>(nlh->lineManager->getLine());
                endCell = endCell->next;
            }
            else endCell = nlh->neighbors = reinterpret_cast<NeighborList*>(nlh->lineManager->getLine());
        }
        else {
            endCell->next = endCell+1; // next cell entry on the same memory line
            endCell = endCell->next;
        }
        // in all the previous case, we got uinitialized memory => initialize it now
        endCell->next = 0;
        endCell->object = object;
        ++nlh->size;
    }
};

// Functor to build neighbor lists, with an object to avoid
struct ListBuilderAvoid : public ListBuilderNoAvoid {
    void* avoidObject;
    NEIGHAND_INLINE ListBuilderAvoid(NeighborListHolder* _nlh, void* o) : ListBuilderNoAvoid(_nlh), avoidObject(o) {}
    NEIGHAND_INLINE void operator()(void* object) NEIGHAND_ALWAYS_INLINE {
        if (object==avoidObject) return;
        ListBuilderNoAvoid::operator()(object);
    }
};


}


NEIGHAND_INLINE void NeighborListHolder::destroy() {
    while (neighbors) {
        NeighborList* next = neighbors->next;
        // Line manager allocates memory line by line, with line size assumed to be 64 bytes
        // De-allocate a line when the next pointer is in a new line, or is 0-terminal
        if ((reinterpret_cast<uint32_t>(next) & 0x3F)==0)
            // Align the current list cell on the 64-bytes boundary to get the line start
            lineManager->releaseLine(reinterpret_cast<void*>(reinterpret_cast<uint32_t>(neighbors) & 0xFFFFFFC0));
        neighbors = next;
    }
}

NEIGHAND_INLINE NeighborListHolder::~NeighborListHolder() {
    destroy();
}

NEIGHAND_INLINE NeighborListHolder::NeighborListHolder(const NeighborListHolder& original)
// Owns the list: copy semantics
// initialize ntotal & neighbors with 0 is intentional for proper list builder usage
: size(0), neighbors(0), lineManager(original.lineManager) {
    detail::ListBuilderNoAvoid builder(this);
    for (NeighborList* list = original.neighbors; list!=0; list = list->next) builder(list->object);
}

NEIGHAND_INLINE NeighborListHolder& NeighborListHolder::operator=(const NeighborListHolder& original) {
    destroy();
    size = 0;
    detail::ListBuilderNoAvoid builder(this);
    for (NeighborList* list = original.neighbors; list!=0; list = list->next) builder(list->object);
    return *this;
}


#ifdef NEIGHAND_SPECIALIZE
template<> int NeighborhoodHandler<NEIGHAND_SPECIALIZE>::findNeighbors(ObjectProxy* p, FloatType d, NeighborListHolder* holder)
#else
template <int exp2divx, int exp2divy, int exp2divz, bool wrapX, bool wrapY, bool wrapZ, bool layerZ>
NEIGHAND_INLINE int NeighborhoodHandler<NEIGHAND_TEMPLATE_ARGUMENTS>::findNeighbors(ObjectProxy* p, FloatType d, NeighborListHolder* holder)
#endif
{
    holder->destroy();
    holder->size = 0;
    holder->lineManager = &lineManager;
    applyToNeighbors(p,d, detail::ListBuilderAvoid(holder, p->object));
    return holder->size;
}

#ifdef NEIGHAND_SPECIALIZE
template<> int NeighborhoodHandler<NEIGHAND_SPECIALIZE>::findNeighbors(FloatType x, FloatType y, FloatType z, FloatType d, NeighborListHolder* holder)
#else
template <int exp2divx, int exp2divy, int exp2divz, bool wrapX, bool wrapY, bool wrapZ, bool layerZ>
NEIGHAND_INLINE int NeighborhoodHandler<NEIGHAND_TEMPLATE_ARGUMENTS>::findNeighbors(FloatType x, FloatType y, FloatType z, FloatType d, NeighborListHolder* holder)
#endif
{
    holder->destroy();
    holder->size = 0;
    holder->lineManager = &lineManager;
    applyToNeighbors(x,y,z,d, detail::ListBuilderNoAvoid(holder));
    return holder->size;
}

#include "neighand_closest.hpp"

// Find only THE closest neighbor
// This function is close enough to "find the N nearest" so as to justify
//   a common source code for maintenance & no copy/paste error risk
// This function is different enough to justify a separate routine for performance reasons
// See the included file
#define NEIGHAND_CLOSEST_N_EQ_1 1
#include "neighand_closest.hpp"
#undef NEIGHAND_CLOSEST_N_EQ_1


// Again using x/y/z API
#define NEIGHAND_APPLY_XYZ_API 1
#include "neighand_closest.hpp"
#define NEIGHAND_CLOSEST_N_EQ_1 1
#include "neighand_closest.hpp"
#undef NEIGHAND_CLOSEST_N_EQ_1
#undef NEIGHAND_APPLY_XYZ_API


} // end namespace

// cleanup preprocessor helper
#undef NEIGHAND_TEMPLATE_ARGUMENTS
