/*
    Neighand: Neighborhood Handling library

    The goal of this class is to find 3D neighbors efficiently.
    It stores the position of objects in 3D. The user can then ask what
    are the neighbor objects for any given point, within a radius
    (aka locality query, or neighborhood search).

    Read the documentation contained in the article "Spherical
    Indexing for Neighborhood Queries" for full details.

    Usage:

    - Either include the "neighand.h" file directly
      Pro: No link needed, all templates are declared in headers.
      Con: Code may be generated multiple times, which increases the
           final binary size.
      Advice: Use this method if you call neighand from only one of your
              files. This avoids generated code duplication, while keeping
              the no-link advantage and inlining best performance.

    - Or specialize the library with NEIGHAND_SPECIALIZE (see examples in
      the tests directory).
      Pro: Unique copy of generated code, smaller binaries. You can
           additionally specialize and generate different CPP
           for different parameters if need be.
      Con: Have to define a flag, compile a cpp and link to it.
      Advice: Use this method if you need neighand directly, at different
              places in your project, or if you need several independent
              locality regions with different parameters.

    Nicolas Brodu, 2006
    Code released according to the GNU GPL, v2 or above.
*/
#ifndef NEIGHAND_H
#define NEIGHAND_H

// posix_memalign may require the next define to be declared in stdlib on some systems
//#define _XOPEN_SOURCE 600
#include <stdlib.h>
#include <stdint.h>
#include <fenv.h>

#include <vector>
#include <fstream>

namespace neighand {

// This may be useful later on to specialize on a special float type, like a software implementation or a type with protected FPU flags. See for example the streflop library.
// This MUST be a 4-bytes type, IEEE754-compliant.
// Changing this, even to streflop types, is untested yet.
#ifdef NEIGHAND_FLOAT_TYPE
typedef NEIGHAND_FLOAT_TYPE FloatType;
#else
typedef float FloatType;
#endif

/*
Force inlining of some functions.
The compiler may think they are too big, and that's usually perfectly justified.
A too big inlined function is actually less performant than a call, because the call
re-uses the code trace L1 cache whereas the duplication due to inlining does not.
However, in our case, some "big" functions are called within critical neighborhood
loops, and only there. Some other functions are just placed in a separate class so
as to make use of template specialization, and should be inlined too.
*/
#ifdef NEIGHAND_DEBUG
#define NEIGHAND_INLINE
#define NEIGHAND_ALWAYS_INLINE __attribute__ ((__noinline__))
#define NEIGHAND_REALLY_ALWAYS_INLINE __attribute__ ((__noinline__))
#else
#if defined(NEIGHAND_SPECIALIZE) || !defined(__GNUC__)
#define NEIGHAND_ALWAYS_INLINE
#define NEIGHAND_REALLY_ALWAYS_INLINE __attribute__ ((__always_inline__))
#define NEIGHAND_INLINE
#else
#define NEIGHAND_ALWAYS_INLINE __attribute__ ((__always_inline__))
#define NEIGHAND_REALLY_ALWAYS_INLINE __attribute__ ((__always_inline__))
#define NEIGHAND_INLINE inline
#endif
#endif

// Internal data structures
#include "neighand_structures.hpp"

// Memory manager that takes care of cache line reuse and alignment
#include "neighand_line_manager.hpp"

// Partially specialized template helpers for different combinations of wrapping
#include "neighand_helpers.hpp"


/*
    Template arguments:
    - exp2div(x|y|z): 2-exponents for the subdivisions in (x|y|z). Ex: 5 = 32 cells.
                      Note: limitation for exp2divx + exp2divy + exp2divz <= 16
                      Note2: Of course, each one should also be >=1
                      Note3: These conditions are compile-time statically checked.
    - wrap(x|y|z): whether to consider a cyclic world in that dimension or not, that
                   is, wrap one edge with the other. Non-cyclic programs should be
                   designed so as to minimize the number of objects "outside" the
                   main region of interest for best performance.
    - layerZ: Reserved for a future extension. Ignored for now.

    Thread safety: All access to this object should be made within the same thread.
*/
template <int exp2divx, int exp2divy, int exp2divz, bool wrapX, bool wrapY, bool wrapZ, bool layerZ = false>
class NeighborhoodHandler {
public:
COMPILE_TIME_ASSERT(exp2divx>=1)
COMPILE_TIME_ASSERT(exp2divy>=1)
COMPILE_TIME_ASSERT(exp2divz>=1)
COMPILE_TIME_ASSERT(exp2divx+exp2divy+exp2divz<=16)


////////////////////////////    PUBLIC API    ////////////////////////////

//// PART 1: Initialization & destruction

    // Constructor arguments:
    // min(x|y|z) is the corner of the region of interest in world coordinates.
    //            It is especially important in case there is no wrapping
    //            along at least one direction.
    // cellSize is the edge size of the cubic cells (the same for each direction).
    //          Thus, for non-cyclic worlds, min[xyz] + cellSize * 2^exp2div[xyz]
    //          is the corner with maximum coordinate for the region of interest.
    // revert size parameters: Distances expressed in cell size units below which
    //                         the algorithm should revert to the bin-lattice
    //                         implementation. The default is 0.0 so there is
    //                         no reversal (except for distance 0, unlike in doc!).
    //                         You should check that for your particular case this
    //                         is correct, especially if you use distances below
    //                         one cell unit in a non-wrapping world. As an
    //                         indication you probably will use small values
    //                         for high load averages or wrapping worlds.
    //                         You may find the methodComp test program useful.
    // sphereSelectionBias is the weighting factor to give to the query sphere
    //                     estimated volumes. See the documentation. 1.0 means only
    //                     the volumes are checked, no bias. Smaller means the
    //                     sphere method will be used preferentially. Look at plots
    //                     from the methodComp program, and decrease this value if
    //                     you see that the "sphere only" method produces better
    //                     results than the proposed default value: you should then
    //                     give it a bias. The reverse situation holds too.
    // precompFile is the name of a binary file that can be used to cache the
    //             precomputations from one run to the other. Default is empty,
    //             so the precomputations are not cached. This should be a valid file
    //             name. Valid files are read, invalid files are overwritten by a
    //             valid one. The file is ONLY used for initialization, and not
    //             accessed anymore after the constructor returns.
    NeighborhoodHandler(FloatType minx, FloatType miny, FloatType minz, FloatType cellSize, FloatType queryRevertSizeCellUnit = 0.0f, FloatType closestRevertSizeCellUnit = 0.0f, FloatType sphereSelectionBias = 1.0f, const char* precompFile = "");

    // convenience constructor that takes only the precompFile as argument, all other parameters
    // are set to their default values
    NeighborhoodHandler(FloatType minx, FloatType miny, FloatType minz, FloatType cellSize, const char* precompFile);

    ~NeighborhoodHandler();


//// PART 2: Managing user objects in the region of interest

    // Insert a user object at the given location (world coordinates)
    // The object can then be considered in the neighborhood queries.
    // Return a proxy = a key identifier for this object.
    // Note: This function maximizes memory re-use in case many objects are inserted/removed
    NEIGHAND_INLINE ObjectProxy* insert(FloatType x, FloatType y, FloatType z, void* object) NEIGHAND_ALWAYS_INLINE;

    // Remove an object proxy from the neighborhood considerations.
    // The proxy must have been returned by "insert" and must be removed only once.
    // Of course, the user object associated with the proxy is completely left unchanged.
    NEIGHAND_INLINE void remove(ObjectProxy* proxy) NEIGHAND_ALWAYS_INLINE;

    // Update a proxy position
    NEIGHAND_INLINE void update(FloatType x, FloatType y, FloatType z, ObjectProxy* proxy) NEIGHAND_ALWAYS_INLINE;


//// PART 3: Neighborhood query functions

    // Main query routines: apply a functor to all neighbors within a given range.
    // The function that takes a proxy as argument excludes that proxy object from the search.
    // The function that takes a x/y/z center position applies to functor to all objects around that center
    // The rationale is that you can use the x/y/z function with p->x,p->y,p->z to include p->object if need be.
    // Arguments:
    // p: An object proxy that is the center of the query
    // x/y/z: The center coordinates
    // d: The maximum distance to look for neighbors (consider using applyToAll for best performances when covering the whole region of interest).
    // f: A functor taking a void* object as argument. This functor is called exactly once for each neighbor.
    template<typename Functor> NEIGHAND_INLINE Functor applyToNeighbors(ObjectProxy* p, FloatType d, Functor f) NEIGHAND_ALWAYS_INLINE;
    template<typename Functor> NEIGHAND_INLINE Functor applyToNeighbors(FloatType x, FloatType y, FloatType z, FloatType d, Functor f) NEIGHAND_ALWAYS_INLINE;

    // Apply a functor to all managed objects, whatever their position
    template<typename Functor> NEIGHAND_INLINE Functor applyToAll(Functor f) NEIGHAND_ALWAYS_INLINE;

    // C API equivalent functions, using a callback instead of a functor.
    typedef void (*Callback)(void* object, void* userData);
    NEIGHAND_INLINE void applyToNeighbors(ObjectProxy* p, FloatType d, Callback f, void* userData) NEIGHAND_ALWAYS_INLINE;
    NEIGHAND_INLINE void applyToNeighbors(FloatType x, FloatType y, FloatType z, FloatType d, Callback f, void* userData) NEIGHAND_ALWAYS_INLINE;
    NEIGHAND_INLINE void applyToAll(Callback f, void* userData) NEIGHAND_ALWAYS_INLINE;

    // Return a list of all neighbors within a given radius.
    // The function that takes a proxy as argument excludes that proxy object from the returned list.
    // The function that takes a x/y/z position returns all objects in neighborhood
    // This is just a wrapper around the previous function with a functor that builds the return list.
    // Note: The list this function returns is NOT sorted by distance.
    //       The distance for each object in that list is not even returned simply because
    //       it needs not be computed in many cases (and thus it would be a waste of time).
    //       => You may still compute it easily if you need it with the provided squaredDistance function
    // Note: The holder object owns and manages the neighbors list. It can be deep copied.
    // ex: NeighborListHolder holder; int nfound = findNeighbors(my_proxy, dist, &holder);
    NEIGHAND_INLINE int findNeighbors(ObjectProxy* p, FloatType d, NeighborListHolder* holder) NEIGHAND_ALWAYS_INLINE;
    NEIGHAND_INLINE int findNeighbors(FloatType x, FloatType y, FloatType z, FloatType d, NeighborListHolder* holder) NEIGHAND_ALWAYS_INLINE;

    // Find only the N closest objects within a given radius.
    // The function that takes a proxy as argument excludes that proxy object from the returned list.
    // The neighbors are returned in the array provided by the user, sorted by distance.
    // Return the number of neighbors found (potentially 0).
    // ex: ClosestNeighbor array[5]; int nfound = findClosestNeighbors(my_position, dist, array, 5);
    NEIGHAND_INLINE int findClosestNeighbors(ObjectProxy* p, FloatType d, ClosestNeighbor* neighbor, unsigned int N) NEIGHAND_ALWAYS_INLINE;
    NEIGHAND_INLINE int findClosestNeighbors(FloatType x, FloatType y, FloatType z, FloatType d, ClosestNeighbor* neighbor, unsigned int N) NEIGHAND_ALWAYS_INLINE;

    // Find only THE closest neighbor within a certain radius
    // The function that takes a proxy as argument excludes that proxy object for a neighbor candidate.
    // Returns 1 if a neighbor was found, 0 otherwise.
    // Note: This function is more performant than calling the previous one with N=1
    // ex: ClosestNeighbor neighbor; int found = findClosestNeighbor(my_position, dist, &neighbor);
    NEIGHAND_INLINE int findClosestNeighbor(ObjectProxy* p, FloatType d, ClosestNeighbor* neighbor) NEIGHAND_ALWAYS_INLINE;
    NEIGHAND_INLINE int findClosestNeighbor(FloatType x, FloatType y, FloatType z, FloatType d, ClosestNeighbor* neighbor) NEIGHAND_ALWAYS_INLINE;



//// PART 4: Helpers and other utilities


    // Return a squared distance for a position difference, taking into account world wrapping.
    // So, this just returns "dx*dx + dy*dy + dz*dz" when there is no wrapping, but
    // performs modulo arithmetic first in the wrapping dimensions.
    // Ex: FloatType dsq = squaredDistance(x1-x2, y1-y2, z1-z2);
    // Note: Exposed to user API from wrap helper. You may find other helper functions useful too.
    // Note2: See neighand_helpers.h for a fast modulo function you may reuse.
    NEIGHAND_INLINE FloatType squaredDistance(FloatType dx, FloatType dy, FloatType dz) NEIGHAND_ALWAYS_INLINE {
        return helper.squaredDistance(dx,dy,dz);
    }

    // Main helper object. This allows to keep a generic template version here,
    // and only the helper will be specialized for the different wrap modes
    WrapHelper<exp2divx,exp2divy,exp2divz,wrapX,wrapY,wrapZ,layerZ> helper;

    // This is the maximum squared distance in number of cells that may separate
    // 2 points in the region of interest
    // Propagated from helper to this name space for convenience
    enum {
        MaxDQ = WrapHelper<exp2divx,exp2divy,exp2divz,wrapX,wrapY,wrapZ,layerZ>::MaxDQ
    };



    // Change the sphere volume weighting.
    // This may be used by your application to switch between different optimal
    // weight factors in real-time, for example when optimizing for specific
    // query distances.
    NEIGHAND_INLINE void setSphereSelectionBias(FloatType sphereSelectionBias) NEIGHAND_ALWAYS_INLINE {
        this->sphereSelectionBias = sphereSelectionBias;
        // use the weighting information
        for (uint32_t i=0; i<=maxWorldDist32; ++i) {
            distanceVolumeTable[i] = uint32_t(distanceVolumeTableBase[i] * sphereSelectionBias);
        }
    }

////////////////////////////    INTERNALS    ////////////////////////////
//protected:

    NEIGHAND_INLINE void removeNoDealloc(ObjectProxy* proxy) NEIGHAND_ALWAYS_INLINE;
    NEIGHAND_INLINE void insertNoAllocate(FloatType x, FloatType y, FloatType z, ObjectProxy* proxy, uint32_t idx) NEIGHAND_ALWAYS_INLINE;

    // Handle lines. Could be made a singleton, though it'd be pretty useless in case only
    // one NeighborhoodHandler object is used, and would then require overhead for thread
    // protection when 2 objects access the singleton from different threads.
    LineManager lineManager;

    // Distance -> sphere cell offsets table lookup
    // indexed by dq, max entries into sphereOffsets array
    uint32_t* baseDistanceArray;
    // all sphere offsets as for no shaving, ordered starting from the center
    uint32_t* sphereOffsets;
    // complements that are not shaved off. Each shavingComplement[x] is an array indexed by dq,
    // of indices into shavingOffsets[x]
    uint32_t* shavingComplement[64];
    // arrays of all cell offset complements, stacked up: complement for dq=0, then for 1, then for 2...
    uint32_t* shavingOffsets[64];

    // Average "sphere" volume for each 1/32th of cell distance
    // allows quick estimation for which method is fastest
    uint32_t maxWorldDist32;
    uint32_t* distanceVolumeTable;
    FloatType* distanceVolumeTableBase;

    // Array of all the cells, a.k.a "blocks", or "bricks"
    CellEntry* cells;
    // Faster to lookup when not using non-empty list
    ObjectProxy** cachedLists;

    // The inverse of the cell size
    FloatType cellSizeSqInv;
    FloatType sphereSelectionBias, closestRevertSize, queryRevertSize;

    // Total number of non-empty cells in the main region
    uint32_t totalNonEmpty;
    // Linked list of objects in the main region
    CellEntry* firstNonEmptyCell;

#ifdef NEIGHAND_SELECT_METHOD_STAT
public:
    uint32_t statSphere, statParallelepiped, statNEList;
    NEIGHAND_INLINE void resetStat() NEIGHAND_ALWAYS_INLINE {
        statSphere = statParallelepiped = statNEList = 0;
    }
#endif
}
#ifdef __GNUC__
__attribute__ ((aligned (64))) // align the neighbor handler itself on cache line size
#endif
;

} // end namespace

// If no explicit specialization is requested, then declare all the routines
// in the header => no link.
#ifndef NEIGHAND_SPECIALIZE
#include "neighand.cpp"
#else

namespace neighand {
// Things that must be declared only once in case of specialization

// ARGL! Fully specializing both base class and member is possible: template<> template<> void Base<A>::member<B>() {}
//       Keeping both generic is possible: template<typename A> template<typename B> void Base<A>::member() {}
//       Specializing only the base class is NOT ALLOWED: template<> template<typename B> void Base<A>::member() {}
//       and in our case wouldn't help with the CPP code separation anyway.
// So, if user can specialize for a functor, fine. Otherwise, hack the generic version so as to be able to compile a generic one in the CPP
template<typename Functor> static void functorToCallback(void* object, void* userData) {
    (*reinterpret_cast<Functor*>(userData))(object);
}
// A generic non-templetized functor that calls the specialized callback
struct CallbackToFunctor {
    typedef void (*Callback)(void* object, void* userData);
    Callback callback;
    void* userData;
    NEIGHAND_INLINE CallbackToFunctor(Callback c, void* data) : callback(c), userData(data) {}
    // And now the functor
    NEIGHAND_INLINE void operator()(void* object) NEIGHAND_REALLY_ALWAYS_INLINE {
        (*callback)(object, userData);
    }
};

// Complete specialization => in the CPP
template<> template<> CallbackToFunctor NeighborhoodHandler<NEIGHAND_SPECIALIZE>::applyToNeighbors<CallbackToFunctor>(ObjectProxy* p, FloatType d, CallbackToFunctor f);
template<> template<> CallbackToFunctor NeighborhoodHandler<NEIGHAND_SPECIALIZE>::applyToNeighbors<CallbackToFunctor>(FloatType x, FloatType y, FloatType z, FloatType d, CallbackToFunctor f);

template<> template<> CallbackToFunctor NeighborhoodHandler<NEIGHAND_SPECIALIZE>::applyToAll(CallbackToFunctor f);

template <int exp2divx, int exp2divy, int exp2divz, bool wrapX, bool wrapY, bool wrapZ, bool layerZ>
template<typename Functor>
NEIGHAND_INLINE Functor NeighborhoodHandler<exp2divx,exp2divy,exp2divz,wrapX,wrapY,wrapZ,layerZ>::applyToNeighbors(ObjectProxy* p, FloatType d, Functor f) {
    applyToNeighbors(p, d, CallbackToFunctor(&functorToCallback<Functor>, &f));
    return f;
}
template <int exp2divx, int exp2divy, int exp2divz, bool wrapX, bool wrapY, bool wrapZ, bool layerZ>
template<typename Functor>
NEIGHAND_INLINE Functor NeighborhoodHandler<exp2divx,exp2divy,exp2divz,wrapX,wrapY,wrapZ,layerZ>::applyToNeighbors(FloatType x, FloatType y, FloatType z, FloatType d, Functor f) {
    applyToNeighbors(x,y,z, d, CallbackToFunctor(&functorToCallback<Functor>, &f));
    return f;
}
template <int exp2divx, int exp2divy, int exp2divz, bool wrapX, bool wrapY, bool wrapZ, bool layerZ>
template<typename Functor>
NEIGHAND_INLINE Functor NeighborhoodHandler<exp2divx,exp2divy,exp2divz,wrapX,wrapY,wrapZ,layerZ>::applyToAll(Functor f) {
    applyToAll(CallbackToFunctor(&functorToCallback<Functor>, &f));
    return f;
}

}
#endif

// End of header protection
#endif

// Specialize for the given set of arguments, implementation is in CPP
// This is expressely outside the multiple inclusion protection, in case
// of multiple specializations
#ifdef NEIGHAND_SPECIALIZE
namespace neighand {
template<> NeighborhoodHandler<NEIGHAND_SPECIALIZE>::NeighborhoodHandler(FloatType minx, FloatType miny, FloatType minz, FloatType cellSize, FloatType queryRevertSizeCellUnit, FloatType closestRevertSizeCellUnit, FloatType sphereSelectionBias, const char* precompFile);
template<> NeighborhoodHandler<NEIGHAND_SPECIALIZE>::NeighborhoodHandler(FloatType minx, FloatType miny, FloatType minz, FloatType cellSize, const char* precompFile);
template<> NeighborhoodHandler<NEIGHAND_SPECIALIZE>::~NeighborhoodHandler();
template<> void NeighborhoodHandler<NEIGHAND_SPECIALIZE>::removeNoDealloc(ObjectProxy* proxy);
template<> void NeighborhoodHandler<NEIGHAND_SPECIALIZE>::insertNoAllocate(FloatType x, FloatType y, FloatType z, ObjectProxy* proxy, uint32_t idx);
template<> ObjectProxy* NeighborhoodHandler<NEIGHAND_SPECIALIZE>::insert(FloatType x, FloatType y, FloatType z, void* object);
template<> void NeighborhoodHandler<NEIGHAND_SPECIALIZE>::remove(ObjectProxy* proxy);
template<> void NeighborhoodHandler<NEIGHAND_SPECIALIZE>::update(FloatType x, FloatType y, FloatType z, ObjectProxy* proxy);
template<> int NeighborhoodHandler<NEIGHAND_SPECIALIZE>::findNeighbors(ObjectProxy* p, FloatType d, NeighborListHolder* holder);
template<> int NeighborhoodHandler<NEIGHAND_SPECIALIZE>::findClosestNeighbors(ObjectProxy* p, FloatType d, ClosestNeighbor* neighbor, unsigned int N);
template<> int NeighborhoodHandler<NEIGHAND_SPECIALIZE>::findClosestNeighbor(ObjectProxy* p, FloatType d, ClosestNeighbor* neighbor);
template<> int NeighborhoodHandler<NEIGHAND_SPECIALIZE>::findNeighbors(FloatType x, FloatType y, FloatType z, FloatType d, NeighborListHolder* holder);
template<> int NeighborhoodHandler<NEIGHAND_SPECIALIZE>::findClosestNeighbors(FloatType x, FloatType y, FloatType z, FloatType d, ClosestNeighbor* neighbor, unsigned int N);
template<> int NeighborhoodHandler<NEIGHAND_SPECIALIZE>::findClosestNeighbor(FloatType x, FloatType y, FloatType z, FloatType d, ClosestNeighbor* neighbor);
// C API
template<> void NeighborhoodHandler<NEIGHAND_SPECIALIZE>::applyToNeighbors(ObjectProxy* p, FloatType d, Callback f, void* userData);
template<> void NeighborhoodHandler<NEIGHAND_SPECIALIZE>::applyToNeighbors(FloatType x, FloatType y, FloatType z, FloatType d, Callback f, void* userData);
template<> void NeighborhoodHandler<NEIGHAND_SPECIALIZE>::applyToAll(Callback f, void* userData);
#ifdef NEIGHAND_SPECIALIZE_FUNCTOR
// Complete specialization => in the CPP
template<> template<> NEIGHAND_SPECIALIZE_FUNCTOR NeighborhoodHandler<NEIGHAND_SPECIALIZE>::applyToNeighbors<NEIGHAND_SPECIALIZE_FUNCTOR>(ObjectProxy* p, FloatType d, NEIGHAND_SPECIALIZE_FUNCTOR f);
template<> template<> NEIGHAND_SPECIALIZE_FUNCTOR NeighborhoodHandler<NEIGHAND_SPECIALIZE>::applyToNeighbors<NEIGHAND_SPECIALIZE_FUNCTOR>(FloatType x, FloatType y, FloatType z, FloatType d, NEIGHAND_SPECIALIZE_FUNCTOR f);
template<> template<> NEIGHAND_SPECIALIZE_FUNCTOR NeighborhoodHandler<NEIGHAND_SPECIALIZE>::applyToAll<NEIGHAND_SPECIALIZE_FUNCTOR>(NEIGHAND_SPECIALIZE_FUNCTOR f);
#endif
}
#endif


