/*
    Neighand: Neighborhood Handling library

    The goal of this project is to find 3D neighbors efficiently.
    Read the documentation contained in the article "Spherical
    Indexing for Neighborhood Queries" for full details.

    This file defines the functions for applying a user-defined
    functor or callback to all objects within a given distance.

    Nicolas Brodu, 2006
    Code released according to the GNU GPL, v2 or above.
*/

#if defined(NEIGHAND_C_CALLBACK_API)

#if defined(NEIGHAND_SPECIALIZE)
#ifdef NEIGHAND_APPLY_XYZ_API
// main function redirected to C callback and C callback redirected to this one
template<> void NeighborhoodHandler<NEIGHAND_SPECIALIZE>::applyToNeighbors(FloatType x, FloatType y, FloatType z, FloatType d, Callback f, void* userData)
#else
// main function redirected to C callback and C callback redirected to this one
template<> void NeighborhoodHandler<NEIGHAND_SPECIALIZE>::applyToNeighbors(ObjectProxy* p, FloatType d, Callback f, void* userData)
#endif
#else
#ifdef NEIGHAND_APPLY_XYZ_API
template <int exp2divx, int exp2divy, int exp2divz, bool wrapX, bool wrapY, bool wrapZ, bool layerZ>
NEIGHAND_INLINE void NeighborhoodHandler<NEIGHAND_TEMPLATE_ARGUMENTS>::applyToNeighbors(FloatType x, FloatType y, FloatType z, FloatType d, Callback f, void* userData)
#else
template <int exp2divx, int exp2divy, int exp2divz, bool wrapX, bool wrapY, bool wrapZ, bool layerZ>
NEIGHAND_INLINE void NeighborhoodHandler<NEIGHAND_TEMPLATE_ARGUMENTS>::applyToNeighbors(ObjectProxy* p, FloatType d, Callback f, void* userData)
#endif
#endif

#else

#if defined(NEIGHAND_SPECIALIZE)

#ifdef NEIGHAND_SPECIALIZE_FUNCTOR
#ifdef NEIGHAND_APPLY_XYZ_API
template<> template<> NEIGHAND_SPECIALIZE_FUNCTOR NeighborhoodHandler<NEIGHAND_SPECIALIZE>::applyToNeighbors<NEIGHAND_SPECIALIZE_FUNCTOR>(FloatType x, FloatType y, FloatType z, FloatType d, NEIGHAND_SPECIALIZE_FUNCTOR f)
#else
template<> template<> NEIGHAND_SPECIALIZE_FUNCTOR NeighborhoodHandler<NEIGHAND_SPECIALIZE>::applyToNeighbors<NEIGHAND_SPECIALIZE_FUNCTOR>(ObjectProxy* p, FloatType d, NEIGHAND_SPECIALIZE_FUNCTOR f)
#endif
#else
#ifdef NEIGHAND_APPLY_XYZ_API
// main function redirected to C callback and C callback redirected to this one
template<> template<> CallbackToFunctor NeighborhoodHandler<NEIGHAND_SPECIALIZE>::applyToNeighbors<CallbackToFunctor>(FloatType x, FloatType y, FloatType z, FloatType d, CallbackToFunctor f)
#else
// main function redirected to C callback and C callback redirected to this one
template<> template<> CallbackToFunctor NeighborhoodHandler<NEIGHAND_SPECIALIZE>::applyToNeighbors<CallbackToFunctor>(ObjectProxy* p, FloatType d, CallbackToFunctor f)
#endif
#endif

#else
#ifdef NEIGHAND_APPLY_XYZ_API
template <int exp2divx, int exp2divy, int exp2divz, bool wrapX, bool wrapY, bool wrapZ, bool layerZ>
template<typename Functor>
NEIGHAND_INLINE Functor NeighborhoodHandler<NEIGHAND_TEMPLATE_ARGUMENTS>::applyToNeighbors(FloatType x, FloatType y, FloatType z, FloatType d, Functor f)
#else
template <int exp2divx, int exp2divy, int exp2divz, bool wrapX, bool wrapY, bool wrapZ, bool layerZ>
template<typename Functor>
NEIGHAND_INLINE Functor NeighborhoodHandler<NEIGHAND_TEMPLATE_ARGUMENTS>::applyToNeighbors(ObjectProxy* p, FloatType d, Functor f)
#endif
#endif

#endif

{

#ifndef NEIGHAND_APPLY_XYZ_API
    FloatType x;
    FloatType y;
    FloatType z;
    *reinterpret_cast<uint32_t*>(&x) = *reinterpret_cast<uint32_t*>(&p->x);
    *reinterpret_cast<uint32_t*>(&y) = *reinterpret_cast<uint32_t*>(&p->y);
    *reinterpret_cast<uint32_t*>(&z) = *reinterpret_cast<uint32_t*>(&p->z);

    void* avoidObject = p->object;
#endif

    helper.clearOutsideFlag(); // inlined to nothing in case of all wraps

    int32_t mincellx, mincelly, mincellz, maxcellx, maxcelly, maxcellz;
    int32_t mincellx_unbounded, mincelly_unbounded, mincellz_unbounded, maxcellx_unbounded, maxcelly_unbounded, maxcellz_unbounded;
    int32_t cellx, celly, cellz;
    uint32_t pvolume, netotal;
    FloatType xcenter, ycenter, zcenter;

    FloatType dsq = d*d;

    // revert to bin-lattice for small distances
    if (d<=queryRevertSize) {
        FloatType d_cellSpace = d * helper.cellSizeInv;

        uint_fast32_t centerCellIndexPacked = helper.getCellIndexForWorldPosition(x,y,z, cellx, celly, cellz, xcenter, ycenter, zcenter);
        if (WrapHelper<NEIGHAND_TEMPLATE_ARGUMENTS>::isInside(centerCellIndexPacked)) {
            pvolume = helper.getInternalParallelepipedAndFlagOutside(mincellx, mincelly, mincellz, maxcellx, maxcelly, maxcellz, mincellx_unbounded, mincelly_unbounded, mincellz_unbounded, maxcellx_unbounded, maxcelly_unbounded, maxcellz_unbounded, xcenter, ycenter, zcenter, d_cellSpace);
        } else {
            pvolume = helper.getIntersectParallelepiped(mincellx, mincelly, mincellz, maxcellx, maxcelly, maxcellz, x, y, z, d);
            helper.flagOutside();
        }

        // Run through the external region list only if necessary
        if (helper.outsideIsFlagged()) {
            register ObjectProxy* plist = helper.getExternalRegionObjectList();
            #include "neighand_apply_processlines_cond.hpp"
        }
    } else {

/* Select faster method

if ((a<=b) && (a<=c)) {
    // a
    return
}

if (b<=c) {
    // b
    return
}

// c

a = weighted sphere estimated volume
b = non-empty list
c = bin-lattice parallelepiped

*/

#ifdef NEIGHAND_APPLY_XYZ_API
    uint_fast32_t centerCellIndexPacked = helper.getCellIndexForWorldPosition(x,y,z, cellx, celly, cellz, xcenter, ycenter, zcenter);
    if (WrapHelper<NEIGHAND_TEMPLATE_ARGUMENTS>::isInside(centerCellIndexPacked)) {
#else
    WrapHelper<NEIGHAND_TEMPLATE_ARGUMENTS>::packedIndexToXYZ(p->cellIndex, cellx, celly, cellz);
    xcenter = (x - helper.minx) * helper.cellSizeInv;
    ycenter = (y - helper.miny) * helper.cellSizeInv;
    zcenter = (z - helper.minz) * helper.cellSizeInv;
    if (WrapHelper<NEIGHAND_TEMPLATE_ARGUMENTS>::isInside(p->cellIndex)) {
#endif

        FloatType d_cellSpace = d * helper.cellSizeInv;

        netotal = totalNonEmpty;
        pvolume = helper.getInternalParallelepipedAndFlagOutside(mincellx, mincelly, mincellz, maxcellx, maxcelly, maxcellz, mincellx_unbounded, mincelly_unbounded, mincellz_unbounded, maxcellx_unbounded, maxcelly_unbounded, maxcellz_unbounded, xcenter, ycenter, zcenter, d_cellSpace);

        uint32_t volTableEntry = uint32_t(d_cellSpace*32.f);
        if (volTableEntry > maxWorldDist32) volTableEntry = maxWorldDist32;
        uint32_t svolume = distanceVolumeTable[volTableEntry];

        if ((svolume <= netotal) && (svolume <= pvolume)) {

#ifdef NEIGHAND_SELECT_METHOD_STAT
            ++statSphere;
#endif

#ifdef NEIGHAND_APPLY_XYZ_API
            uint_fast32_t centerCellIndex = WrapHelper<NEIGHAND_TEMPLATE_ARGUMENTS>::unpack(centerCellIndexPacked);
#else
            uint_fast32_t centerCellIndex = WrapHelper<NEIGHAND_TEMPLATE_ARGUMENTS>::unpack(p->cellIndex);
#endif

            uint32_t dmax = uint32_t(d_cellSpace);
            FloatType dsq_cellSpace = d_cellSpace*d_cellSpace;
            uint32_t lastdq = uint32_t(dsq_cellSpace);//dsq*cellSizeSqInv);
            uint32_t dmax_p1 = dmax + 1;

            // See documentation article: select one of the 64 specialized array
            uint32_t shaving = ((cellx - mincellx_unbounded - dmax_p1) >> 31)
            | (((maxcellx_unbounded - cellx - dmax_p1) & 0x80000000) >> 30)
            | (((celly - mincelly_unbounded - dmax_p1) & 0x80000000) >> 29)
            | (((maxcelly_unbounded - celly - dmax_p1) & 0x80000000) >> 28)
            | (((cellz - mincellz_unbounded - dmax_p1) & 0x80000000) >> 27)
            | (((maxcellz_unbounded - cellz - dmax_p1) & 0x80000000) >> 26);

            int32_t dmaxSq_m1 = dmax*dmax - 1;

            FloatType dm1 = d_cellSpace - WrapHelper<NEIGHAND_TEMPLATE_ARGUMENTS>::distCompHeuristicConstant();
            int32_t dm1sq = uint32_t(dm1*dm1);
            if (dm1<FloatType(0.f)) dm1sq = -1;

            if (lastdq>MaxDQ) {
                lastdq = MaxDQ;
                dmaxSq_m1 = MaxDQ;
                dm1sq = MaxDQ;
                shaving = 0;
            }

            uint32_t* baseMaxUp = &sphereOffsets[baseDistanceArray[dmaxSq_m1]];

            uint32_t* interOffsetList = &shavingOffsets[shaving][shavingComplement[shaving][dmaxSq_m1]];
            uint32_t* interMaxUp = &shavingOffsets[shaving][shavingComplement[shaving][lastdq]];

            xcenter -= cellx;
            ycenter -= celly;
            zcenter -= cellz;

            uint32_t dbldmax = dmax_p1 + dmax_p1 + 1;
            uint32_t dbldmax3 = dbldmax+dbldmax+dbldmax;
            uint32_t all_cache[dbldmax3];
            for (uint32_t i=0; i<dbldmax3; ++i) all_cache[i] = 0;
            // alias to mid-array to allow indexing by negative offsets
            FloatType* cachedDeltaX = reinterpret_cast<FloatType*>(all_cache) + dmax_p1;
            FloatType* cachedDeltaY = cachedDeltaX + dbldmax;
            FloatType* cachedDeltaZ = cachedDeltaY + dbldmax;

            uint32_t* baseOffsetList = &sphereOffsets[0];
            if (d_cellSpace > FloatType(1.73205081)) {
                d_cellSpace -= FloatType(1.73205081);
                baseOffsetList = &sphereOffsets[uint32_t(d_cellSpace * d_cellSpace)];
                register uint32_t* offsetlist = &sphereOffsets[0];
                while (offsetlist < baseOffsetList) {
                    uint32_t offset = *offsetlist++;
                    uint32_t unpackedCell = centerCellIndex + offset;
                    if (WrapHelper<NEIGHAND_TEMPLATE_ARGUMENTS>::isUnpackedOutside(unpackedCell)) continue;
                    register ObjectProxy* plist = cachedLists[WrapHelper<NEIGHAND_TEMPLATE_ARGUMENTS>::packwrap(unpackedCell)];
                    #define NEIGHAND_APPLY_CHECK_FIRST
                    #include "neighand_apply_processlines_uncond.hpp"
                    #undef NEIGHAND_APPLY_CHECK_FIRST
                }
            }

            register uint32_t* offsetlist = baseOffsetList;
            baseOffsetList = &sphereOffsets[baseDistanceArray[dm1sq]];
            while (offsetlist < baseOffsetList) {
                uint32_t offset = *offsetlist++;
                uint32_t unpackedCell = centerCellIndex + offset;
                if (WrapHelper<NEIGHAND_TEMPLATE_ARGUMENTS>::isUnpackedOutside(unpackedCell)) continue;
                register ObjectProxy* plist = cachedLists[WrapHelper<NEIGHAND_TEMPLATE_ARGUMENTS>::packwrap(unpackedCell)];
                #define NEIGHAND_APPLY_CHECK_FIRST
                #include "neighand_apply_processlines_cond.hpp"
                #undef NEIGHAND_APPLY_CHECK_FIRST
            }

            while (offsetlist < baseMaxUp) {
                uint32_t offset = *offsetlist++;
                uint32_t unpackedCell = centerCellIndex + offset;
                if (WrapHelper<NEIGHAND_TEMPLATE_ARGUMENTS>::isUnpackedOutside(unpackedCell)) continue;
                #include "neighand_apply_checkdist.hpp"
                register ObjectProxy* plist = cachedLists[WrapHelper<NEIGHAND_TEMPLATE_ARGUMENTS>::packwrap(unpackedCell)];
                #define NEIGHAND_APPLY_CHECK_FIRST
                #include "neighand_apply_processlines_cond.hpp"
                #undef NEIGHAND_APPLY_CHECK_FIRST
            }

            offsetlist = interOffsetList;
            while (offsetlist < interMaxUp) {
                uint32_t offset = *offsetlist++;
                uint32_t unpackedCell = centerCellIndex + offset;
                if (WrapHelper<NEIGHAND_TEMPLATE_ARGUMENTS>::isUnpackedOutside(unpackedCell)) continue;
                #include "neighand_apply_checkdist.hpp"
                register ObjectProxy* plist = cachedLists[WrapHelper<NEIGHAND_TEMPLATE_ARGUMENTS>::packwrap(unpackedCell)];
                #define NEIGHAND_APPLY_CHECK_FIRST
                #include "neighand_apply_processlines_cond.hpp"
                #undef NEIGHAND_APPLY_CHECK_FIRST
            }

            // Now run through the external region list only if necessary
            // inlined to false in all-wrap case
            if (helper.outsideIsFlagged()) {
                register ObjectProxy* plist = helper.getExternalRegionObjectList();
                #include "neighand_apply_processlines_cond.hpp"
            }

#ifdef NEIGHAND_C_CALLBACK_API
            return;
#else
            return f;
#endif
        } // end of sphere selection

    // Query point is outside. Sphere table lookup cannot be used directly (in some cases it could)
    } else {
        pvolume = helper.getIntersectParallelepiped(mincellx, mincelly, mincellz, maxcellx, maxcelly, maxcellz, x, y, z, d);
        netotal = totalNonEmpty;
        helper.flagOutside();
    }

    if (helper.outsideIsFlagged()) {
        register ObjectProxy* plist = helper.getExternalRegionObjectList();
        #include "neighand_apply_processlines_cond.hpp"
    }

    // non-empty list is shorter
    if (netotal < pvolume) {
#ifdef NEIGHAND_SELECT_METHOD_STAT
        ++statNEList;
#endif
        for (CellEntry* entry = firstNonEmptyCell; entry!=0; entry = entry->nextNonEmpty) {
            register ObjectProxy* plist = entry->objects;
            #include "neighand_apply_processlines_cond.hpp"
        }
#ifdef NEIGHAND_C_CALLBACK_API
        return;
#else
        return f;
#endif
    }

#ifdef NEIGHAND_SELECT_METHOD_STAT
    ++statParallelepiped;
#endif

    } // end of small distance reversion to bin-lattice

    // Whole world? faster than triple-loop
    if (pvolume==WrapHelper<NEIGHAND_TEMPLATE_ARGUMENTS>::MaxVolume) {
        ObjectProxy** cList = cachedLists;
        ObjectProxy** lastList = cachedLists + WrapHelper<NEIGHAND_TEMPLATE_ARGUMENTS>::MaxVolume;
        while (cList!=lastList) {
            register ObjectProxy* plist = *cList++;
            #define NEIGHAND_APPLY_CHECK_FIRST
            #include "neighand_apply_processlines_cond.hpp"
            #undef NEIGHAND_APPLY_CHECK_FIRST
        }
#ifdef NEIGHAND_C_CALLBACK_API
        return;
#else
        return f;
#endif
    }

    for (int_fast32_t cellz = mincellz; cellz <= maxcellz; ++cellz) {
        for (int_fast32_t celly = mincelly; celly <= maxcelly; ++celly) {
            for (int_fast32_t cellx = mincellx; cellx <= maxcellx; ++cellx) {
                register ObjectProxy* plist = cachedLists[WrapHelper<NEIGHAND_TEMPLATE_ARGUMENTS>::xyzToPackedIndex(cellx,celly,cellz)];
                #define NEIGHAND_APPLY_CHECK_FIRST
                #include "neighand_apply_processlines_cond.hpp"
                #undef NEIGHAND_APPLY_CHECK_FIRST
            }
        }
    }

#ifndef NEIGHAND_C_CALLBACK_API
    return f;
#endif
}
