/*
    Neighand: Neighborhood Handling library

    The goal of this project is to find 3D neighbors efficiently.
    Read the documentation contained in the article "Spherical
    Indexing for Neighborhood Queries" for full details.

    This file defines the functions for finding only the N
    nearest neighbors, within a given distance.

    Nicolas Brodu, 2006
    Code released according to the GNU GPL, v2 or above.
*/

#ifdef NEIGHAND_APPLY_XYZ_API

#if defined(NEIGHAND_SPECIALIZE)

#if defined(NEIGHAND_CLOSEST_N_EQ_1)
template<> int NeighborhoodHandler<NEIGHAND_SPECIALIZE>::findClosestNeighbor(FloatType x, FloatType y, FloatType z, FloatType d, ClosestNeighbor* neighbor)
#else
template<> int NeighborhoodHandler<NEIGHAND_SPECIALIZE>::findClosestNeighbors(FloatType x, FloatType y, FloatType z, FloatType d, ClosestNeighbor* neighbor, unsigned int N)
#endif

#else

#if defined(NEIGHAND_CLOSEST_N_EQ_1)
template <int exp2divx, int exp2divy, int exp2divz, bool wrapX, bool wrapY, bool wrapZ, bool layerZ>
NEIGHAND_INLINE int NeighborhoodHandler<NEIGHAND_TEMPLATE_ARGUMENTS>::findClosestNeighbor(FloatType x, FloatType y, FloatType z, FloatType d, ClosestNeighbor* neighbor)
#else
template <int exp2divx, int exp2divy, int exp2divz, bool wrapX, bool wrapY, bool wrapZ, bool layerZ>
NEIGHAND_INLINE int NeighborhoodHandler<NEIGHAND_TEMPLATE_ARGUMENTS>::findClosestNeighbors(FloatType x, FloatType y, FloatType z, FloatType d, ClosestNeighbor* neighbor, unsigned int N)
#endif

#endif

// else for X/Y/Z API
#else

#if defined(NEIGHAND_SPECIALIZE)

#if defined(NEIGHAND_CLOSEST_N_EQ_1)
template<> int NeighborhoodHandler<NEIGHAND_SPECIALIZE>::findClosestNeighbor(ObjectProxy* p, FloatType d, ClosestNeighbor* neighbor)
#else
template<> int NeighborhoodHandler<NEIGHAND_SPECIALIZE>::findClosestNeighbors(ObjectProxy* p, FloatType d, ClosestNeighbor* neighbor, unsigned int N)
#endif

#else

#if defined(NEIGHAND_CLOSEST_N_EQ_1)
template <int exp2divx, int exp2divy, int exp2divz, bool wrapX, bool wrapY, bool wrapZ, bool layerZ>
NEIGHAND_INLINE int NeighborhoodHandler<NEIGHAND_TEMPLATE_ARGUMENTS>::findClosestNeighbor(ObjectProxy* p, FloatType d, ClosestNeighbor* neighbor)
#else
template <int exp2divx, int exp2divy, int exp2divz, bool wrapX, bool wrapY, bool wrapZ, bool layerZ>
NEIGHAND_INLINE int NeighborhoodHandler<NEIGHAND_TEMPLATE_ARGUMENTS>::findClosestNeighbors(ObjectProxy* p, FloatType d, ClosestNeighbor* neighbor, unsigned int N)
#endif

#endif

#endif

{

#ifndef NEIGHAND_APPLY_XYZ_API
    FloatType x;
    FloatType y;
    FloatType z;
    *reinterpret_cast<uint32_t*>(&x) = *reinterpret_cast<uint32_t*>(&p->x);
    *reinterpret_cast<uint32_t*>(&y) = *reinterpret_cast<uint32_t*>(&p->y);
    *reinterpret_cast<uint32_t*>(&z) = *reinterpret_cast<uint32_t*>(&p->z);

    void* avoidObject = p->object;
#endif

    FloatType dsq = d*d;

#if defined(NEIGHAND_CLOSEST_N_EQ_1)
    // Initialize dist with max query distance
    FloatType neighborSquaredDistance = dsq;
    // but keep this flag to remind this is not a real object, just a limit
    void* neighborObject = 0;
#else
    // number found
    uint_fast32_t nfound = 0;

    // Initialize dist array with max possible non-infinity distance
    for (unsigned int i=0; i<N; ++i) {
        *reinterpret_cast<uint32_t*>(&neighbor[i].squaredDistance) = uint32_t(0x7F7FFFFF);
        // neighborsHolder[i].object = 0; // not necessary
    }
#endif

    helper.clearOutsideFlag(); // inlined to nothing in case of all wraps

    int32_t mincellx, mincelly, mincellz, maxcellx, maxcelly, maxcellz;
    int32_t mincellx_unbounded, mincelly_unbounded, mincellz_unbounded, maxcellx_unbounded, maxcelly_unbounded, maxcellz_unbounded;
    int32_t cellx, celly, cellz;
    FloatType xcenter, ycenter, zcenter;

#ifdef NEIGHAND_APPLY_XYZ_API
    uint_fast32_t centerCellIndexPacked = helper.getCellIndexForWorldPosition(x,y,z, cellx, celly, cellz, xcenter, ycenter, zcenter);
    if (WrapHelper<NEIGHAND_TEMPLATE_ARGUMENTS>::isInside(centerCellIndexPacked)) {
        uint_fast32_t centerCellIndex = WrapHelper<NEIGHAND_TEMPLATE_ARGUMENTS>::unpack(centerCellIndexPacked);
#else
    WrapHelper<NEIGHAND_TEMPLATE_ARGUMENTS>::packedIndexToXYZ(p->cellIndex, cellx, celly, cellz);
    xcenter = (x - helper.minx) * helper.cellSizeInv;
    ycenter = (y - helper.miny) * helper.cellSizeInv;
    zcenter = (z - helper.minz) * helper.cellSizeInv;
    if (WrapHelper<NEIGHAND_TEMPLATE_ARGUMENTS>::isInside(p->cellIndex)) {
        // Unpacked index for sphere search: this is the center, necessarily inside
        uint_fast32_t centerCellIndex = WrapHelper<NEIGHAND_TEMPLATE_ARGUMENTS>::unpack(p->cellIndex);
#endif

        FloatType d_cellSpace = d * helper.cellSizeInv;

        helper.getInternalParallelepiped(mincellx, mincelly, mincellz, maxcellx, maxcelly, maxcellz, mincellx_unbounded, mincelly_unbounded, mincellz_unbounded, maxcellx_unbounded, maxcelly_unbounded, maxcellz_unbounded, xcenter, ycenter, zcenter, d_cellSpace);

        if (d <= closestRevertSize) {
            for (int_fast32_t cellz = mincellz; cellz <= maxcellz; ++cellz) {
                for (int_fast32_t celly = mincelly; celly <= maxcelly; ++celly) {
                    for (int_fast32_t cellx = mincellx; cellx <= maxcellx; ++cellx) {
                        register ObjectProxy* plist = cachedLists[WrapHelper<NEIGHAND_TEMPLATE_ARGUMENTS>::xyzToPackedIndex(cellx,celly,cellz)];
                        #define NEIGHAND_CLOSEST_CHECK_FIRST
                        #include "neighand_closest_processlines.hpp"
                        #undef NEIGHAND_CLOSEST_CHECK_FIRST
                    }
                }
            }
            helper.flagOutside(x,y,z, d);
            if (helper.outsideIsFlagged()) {
                ObjectProxy* plist = helper.getExternalRegionObjectList();
                #include "neighand_closest_processlines.hpp"
            }
#if defined(NEIGHAND_CLOSEST_N_EQ_1)
            neighbor->object = neighborObject;
            neighbor->squaredDistance = neighborSquaredDistance;
            return neighborObject != 0;
#else
            return nfound;
#endif
        }

        // flag that is set to true as soon as N neighbors are found
        // At that point, possible closer candidates are necessarily inside the
        // cells at the sphere edge for that distance
        // Cells further away than that cannot contain closer neighbors
        bool NWereFound = false;

        uint32_t dmax = uint32_t(d_cellSpace);
        FloatType dsq_cellSpace = d_cellSpace*d_cellSpace;
        uint32_t lastdq = uint32_t(dsq_cellSpace);//dsq*cellSizeSqInv);
        uint32_t dmax_p1 = dmax + 1;

        uint32_t shaving = ((cellx - mincellx_unbounded - dmax_p1) >> 31)
        | (((maxcellx_unbounded - cellx - dmax_p1) & 0x80000000) >> 30)
        | (((celly - mincelly_unbounded - dmax_p1) & 0x80000000) >> 29)
        | (((maxcelly_unbounded - celly - dmax_p1) & 0x80000000) >> 28)
        | (((cellz - mincellz_unbounded - dmax_p1) & 0x80000000) >> 27)
        | (((maxcellz_unbounded - cellz - dmax_p1) & 0x80000000) >> 26);

        int32_t dmaxSq_m1 = dmax*dmax - 1;

        FloatType dm1 = d_cellSpace - WrapHelper<NEIGHAND_TEMPLATE_ARGUMENTS>::distCompHeuristicConstant();
        int32_t dm1sq = uint32_t(dm1*dm1);
        if (dm1<FloatType(0.f)) dm1sq = -1;

        if (lastdq>MaxDQ) {
            lastdq = MaxDQ;
            dmaxSq_m1 = MaxDQ;
            shaving = 0;
        }

        xcenter -= cellx;
        ycenter -= celly;
        zcenter -= cellz;

        uint32_t dbldmax = dmax_p1 + dmax_p1 + 1;
        uint32_t dbldmax3 = dbldmax+dbldmax+dbldmax;
        uint32_t all_cache[dbldmax3];
        for (uint32_t i=0; i<dbldmax3; ++i) all_cache[i] = 0;
        // alias to mid-array to allow indexing by negative offsets
        FloatType* cachedDeltaX = reinterpret_cast<FloatType*>(all_cache) + dmax_p1;
        FloatType* cachedDeltaY = cachedDeltaX + dbldmax;
        FloatType* cachedDeltaZ = cachedDeltaY + dbldmax;

        uint32_t* baseMaxUp = &sphereOffsets[baseDistanceArray[dmaxSq_m1]];
        uint32_t* cmpCellLimit = &sphereOffsets[baseDistanceArray[dm1sq]];

        uint32_t* interOffsetList = &shavingOffsets[shaving][shavingComplement[shaving][dmaxSq_m1]];
        uint32_t* interMaxUp = &shavingOffsets[shaving][shavingComplement[shaving][lastdq]];

        uint32_t* offsetlist = &sphereOffsets[0];
        while (offsetlist < baseMaxUp) {
            uint32_t offset = *offsetlist++;

            uint32_t unpackedCell = centerCellIndex + offset;
            if (WrapHelper<NEIGHAND_TEMPLATE_ARGUMENTS>::isUnpackedOutside(unpackedCell)) continue;
            if (offsetlist<=cmpCellLimit) {
                #include "neighand_apply_checkdist.hpp"
            }

            register ObjectProxy* plist = cachedLists[WrapHelper<NEIGHAND_TEMPLATE_ARGUMENTS>::packwrap(unpackedCell)];

            #define NEIGHAND_CLOSEST_CHECK_FIRST
            #include "neighand_closest_processlines.hpp"
            #undef NEIGHAND_CLOSEST_CHECK_FIRST

            // N neighbors were found ?
            // Potentially closer candidates can only be on the edge of the sphere at that distance
            // Test only valid the first time, or the edge could be pushed up to max!
            // use negative logic to save one branching with continue
            if (NWereFound) continue;
#if defined(NEIGHAND_CLOSEST_N_EQ_1)
            if (neighborObject==0) continue;
#else
            if (nfound != N) continue;
#endif
            NWereFound = true;

            int_fast32_t aboveDQ = uint_fast32_t(
#if defined(NEIGHAND_CLOSEST_N_EQ_1)
                neighborSquaredDistance
#else
                dsq // was updated
#endif
            * cellSizeSqInv);

            // Further neighbors can only be in edge now
            if (aboveDQ <= dmaxSq_m1) {
                baseMaxUp = &sphereOffsets[baseDistanceArray[aboveDQ]];
                interOffsetList = interMaxUp = 0;
            } else if (uint32_t(aboveDQ) < lastdq) {
                interMaxUp = &shavingOffsets[shaving][shavingComplement[shaving][aboveDQ]];
            }
        } // end sphere list run up

        offsetlist = interOffsetList;
        while (offsetlist < interMaxUp) {
            uint32_t offset = *offsetlist++;

            uint32_t unpackedCell = centerCellIndex + offset;
            if (WrapHelper<NEIGHAND_TEMPLATE_ARGUMENTS>::isUnpackedOutside(unpackedCell)) continue;
            if (offsetlist<=cmpCellLimit) {
                #include "neighand_apply_checkdist.hpp"
            }

            register ObjectProxy* plist = cachedLists[WrapHelper<NEIGHAND_TEMPLATE_ARGUMENTS>::packwrap(unpackedCell)];

            #define NEIGHAND_CLOSEST_CHECK_FIRST
            #include "neighand_closest_processlines.hpp"
            #undef NEIGHAND_CLOSEST_CHECK_FIRST

            if (NWereFound) continue;
#if defined(NEIGHAND_CLOSEST_N_EQ_1)
            if (neighborObject==0) continue;
#else
            if (nfound != N) continue;
#endif
            NWereFound = true;
            uint_fast32_t aboveDQ = uint_fast32_t(
#if defined(NEIGHAND_CLOSEST_N_EQ_1)
                neighborSquaredDistance
#else
                dsq // was updated
#endif
            * cellSizeSqInv);

            // Further neighbors can only be in edge now
            if (aboveDQ < lastdq) {
                interMaxUp = &shavingOffsets[shaving][shavingComplement[shaving][aboveDQ]];
            }
        }

        // Now run through the external region list only if necessary
        if (helper.dsqImpliesOutside(x,y,z,
#if defined(NEIGHAND_CLOSEST_N_EQ_1)
        neighborSquaredDistance
#else
        dsq
#endif
        )) {
            // Region is not empty at this point: otherwise it would not have been flagged
            register ObjectProxy* plist = helper.getExternalRegionObjectList();
            #include "neighand_closest_processlines.hpp"
        }

        // may return 0 is there was no neighbor within the required distance
#if defined(NEIGHAND_CLOSEST_N_EQ_1)
        neighbor->object = neighborObject;
        neighbor->squaredDistance = neighborSquaredDistance;
        return neighborObject != 0;
#else
        return nfound;
#endif

    // End of query point inside main region => case for outside
    } else {

        // Query point is outside. Algo:
        // - Run through the external region since we'll have to do it anyway
        // - If the N closest neighbors found are all closer than main region, then it's done
        // - Otherwise process main region, using parallelepiped/list trick

        // First run through the external region list
        if (helper.outsideIsNonEmpty()) {
            ObjectProxy* plist = helper.getExternalRegionObjectList();
            #include "neighand_closest_processlines.hpp"
        }

        // Check if all neighbors found are closer than main region, or if query distance too short => done!
#if defined(NEIGHAND_CLOSEST_N_EQ_1)
        if ((neighborObject!=0) && (helper.getSquaredDistanceToMainRegion(x,y,z) > neighborSquaredDistance)) {
            neighbor->object = neighborObject;
            neighbor->squaredDistance = neighborSquaredDistance;
            return neighborObject != 0;
       }
#else
        if ((nfound == N) && (helper.getSquaredDistanceToMainRegion(x,y,z) > dsq)) // dsq was updated
            return nfound;
#endif

        // So now, we know that an intersection with main region exists and that it must be processed.
        int32_t mincellx, mincelly, mincellz, maxcellx, maxcelly, maxcellz;
        uint_fast32_t pvolume = helper.getIntersectParallelepiped(mincellx, mincelly, mincellz, maxcellx, maxcelly, maxcellz, x, y, z, d);

        // Use non-empty main list only when it's worth it
        if (totalNonEmpty <= pvolume) {
            for (CellEntry* entry = firstNonEmptyCell; entry!=0; entry = entry->nextNonEmpty) {
                register ObjectProxy* plist = entry->objects;
                #include "neighand_closest_processlines.hpp"
            }

            // Done! May return 0 if there was no neighbor within the required distance
#if defined(NEIGHAND_CLOSEST_N_EQ_1)
            neighbor->object = neighborObject;
            neighbor->squaredDistance = neighborSquaredDistance;
            return neighborObject != 0;
#else
            return nfound;
#endif
        }

        if (pvolume==WrapHelper<NEIGHAND_TEMPLATE_ARGUMENTS>::MaxVolume) {
            // Faster than triple-loop
            ObjectProxy** cList = cachedLists;
            ObjectProxy** lastList = cachedLists + WrapHelper<NEIGHAND_TEMPLATE_ARGUMENTS>::MaxVolume;
            while (cList!=lastList) {
                register ObjectProxy* plist = *cList++;
                #define NEIGHAND_CLOSEST_CHECK_FIRST
                #include "neighand_closest_processlines.hpp"
                #undef NEIGHAND_CLOSEST_CHECK_FIRST
            }
        } else {
            for (int_fast32_t cellz = mincellz; cellz <= maxcellz; ++cellz) {
                for (int_fast32_t celly = mincelly; celly <= maxcelly; ++celly) {
                    for (int_fast32_t cellx = mincellx; cellx <= maxcellx; ++cellx) {
                        register ObjectProxy* plist = cachedLists[WrapHelper<NEIGHAND_TEMPLATE_ARGUMENTS>::xyzToPackedIndex(cellx,celly,cellz)];
                        #define NEIGHAND_CLOSEST_CHECK_FIRST
                        #include "neighand_closest_processlines.hpp"
                        #undef NEIGHAND_CLOSEST_CHECK_FIRST
                    }
                }
            }
        }
    }

    // may return 0 if there was no neighbor within the required distance
#if defined(NEIGHAND_CLOSEST_N_EQ_1)
    neighbor->object = neighborObject;
    neighbor->squaredDistance = neighborSquaredDistance;
    return neighborObject != 0;
#else
    return nfound;
#endif
}

