/*
    Neighand: Neighborhood Handling library

    The goal of this project is to find 3D neighbors efficiently.
    Read the documentation contained in the article "Spherical
    Indexing for Neighborhood Queries" for full details.

    This file defines a memory manager.
    This class is reponsible for allocating blocks of memory that fit
    in cache lines (especially alignment conditions).
    It also allows for re-use of recently fred blocks, hence cache lines.

    The assumption is that a cache line is 64 bytes long. Older machines may
    have 32-byte lines, but then these are allocated 2 by 2 thanks to the
    alignment: this will be faster than random alignment that may use 3 lines.

    Nicolas Brodu, 2006
    Code released according to the GNU GPL, v2 or above.
*/
#ifndef NEIGHAND_LINE_MANAGER_HPP
#define NEIGHAND_LINE_MANAGER_HPP

#ifndef NEIGHAND_H
#warning You should not include this file directly unless you know what you're doing. Please use neighand.h
#endif

// allocating & deallocating 64-bytes aligned lines with new, posix_memalign, or malloc is inefficient
// both in terms of speed and memory waste
// => allocate big chunks and then manage lines inside the chunk
// Note: additional bonus: use LIFO instead of FIFO for lines => they may still be in cache
struct LineManager {

    // generic line may actually be 1 true line if line is 64-bytes, or 2 32-byte lines on older PCs
    // In either case, aligning avoids overlapping the structure on different physical lines.
    struct GenericLine {
        uint8_t bytes[64];
    };
    COMPILE_TIME_ASSERT(sizeof(GenericLine)==64)

    // Use chunks of contiguous lines, the goal is to minimize the number of calls to new chunks.
    // A compromise has to be made on fragmentation.
    // Leave one line or 2 for system allocation variables if system is using pages for chunk alloc
    struct Chunk {
        enum {numLine = 254}; // 16K chunk?
        GenericLine lines[numLine];
        Chunk * next; // maintain linked list of allocated chunks. Could be pre-allocated?
    };

    // When a chunk is full, another chunk is selected
    // When all chunks are full, a new chunk is allocated
    // See what happens on destruction below
    GenericLine* freeLine;

    ObjectProxy* freeProxy; // 0

    // store chunk pointers. Growable on demand
    Chunk* chunks;

    LineManager() : freeLine(0), freeProxy(0), chunks(0) {
    }

    ~LineManager() {
        // cleanup chunks
        while (chunks!=0) {
            Chunk* next = chunks->next;
            free(chunks);
            chunks = next;
        }
    }

    // TODO: prevent copy!

    // Generic factory to return ONE line.
    NEIGHAND_INLINE GenericLine* getLine() {
        // No free chunk left (or init case) => allocate one
        // TODO: check for out-of-memory condition
        if (!freeLine) {
            void* newMem;
            // align on 64 bytes
            posix_memalign(&newMem, 64, sizeof(Chunk));
            // define alias
            Chunk* newChunk = reinterpret_cast<Chunk*>(newMem);
            // initialize struct : link all the lines in the new chunk
            for (int i=0; i<Chunk::numLine-1; ++i) {
                // use first bytes as pointer to next line
                GenericLine** linkedLine = reinterpret_cast<GenericLine**>(&newChunk->lines[i]);
                *linkedLine = &newChunk->lines[i+1];
            }
            // end the links
            *reinterpret_cast<GenericLine**>(&newChunk->lines[Chunk::numLine-1]) = 0;
            // use the new lines
            freeLine = &newChunk->lines[0];
            // maintain chunk list
            newChunk->next = chunks;
            chunks = newChunk;
        }

        // position the variable at the right place. Placement new would initialize member types, etc.
        // reinterpret_cast the line does not initialize.
        // This is faster, and care was taken for initial conditions in this project.
        GenericLine* ret = freeLine;

        // prepare next free line (or 0): all free lines are also pointers to the next by construction
        freeLine = *reinterpret_cast<GenericLine**>(freeLine);

        // done
        return ret;
    }

    // TODO: g++ attribute malloc?

    // Free a line. TODO: maintain empty chunks and when there are too many of them, start freeing chunks
    NEIGHAND_INLINE void releaseLine(void* a_line) {
/*        Chunk* a_chunk = 0;
        // run through all chunks and compare line pointer with start/end of chunks
        for (std::vector<Chunk*>::iterator it = chunks.begin(); it!=chunks.end(); ++it) {
            // check if this line belongs to this chunk.
            // mem organization on PC is that stack grows with increasing pointer address: first check is >
            if (reinterpret_cast<GenericLine*>(a_line) > &(*it)->lines[Chunk::numLine-1]) continue;
            // second bound: is this a line for this chunk?
            if (reinterpret_cast<GenericLine*>(a_line) < &(*it)->lines[0]) continue;
            // OK, this chunk now has a line to be freed
            a_chunk = *it;
            break;
        }
        // not one of our lines. error condition, exception?
        if (!a_chunk) return;
*/
        // place a pointer to the next free line in the new free line
        *reinterpret_cast<GenericLine**>(a_line) = freeLine;
        // update free line pointer for next call
        freeLine = reinterpret_cast<GenericLine*>(a_line);
    }


    // Maintain proxies too
    struct ProxyLine {
        enum { NumProxies = 2 };
        ObjectProxy proxy[NumProxies];
    }; // 64
    COMPILE_TIME_ASSERT(sizeof(ProxyLine)==64)

    // get a proxy, allocate line if necessary
    NEIGHAND_INLINE ObjectProxy* getProxy() {
        // same algo, but line => proxy, chunk => line
        if (!freeProxy) {
            ProxyLine* proxyLine = reinterpret_cast<ProxyLine*>(getLine());
            // prepare the line
            for (int i=0; i<ProxyLine::NumProxies-1; ++i) {
                // use first bytes as pointer to next free proxy
                ObjectProxy** linkedProxy = reinterpret_cast<ObjectProxy**>(&proxyLine->proxy[i]);
                *linkedProxy = &proxyLine->proxy[i+1];
                // Use flag for nullity (see neighand_structures)
                proxyLine->proxy[i].cell = 0;
            }
            // end the links
            *reinterpret_cast<ObjectProxy**>(&proxyLine->proxy[ProxyLine::NumProxies-1]) = 0;
            // Use flag for nullity (see neighand_structures)
            proxyLine->proxy[ProxyLine::NumProxies-1].cell = 0;
            // use the new proxy
            freeProxy = &proxyLine->proxy[0];
        }

        // use free object for return
        ObjectProxy* ret = freeProxy;

        // prepare next free proxy (or 0)
        freeProxy = *reinterpret_cast<ObjectProxy**>(freeProxy);

        // done
        return ret;
    }

    // release a proxy. may fragment but only when realeasing memory, and even then care is taken to auto-clean
    NEIGHAND_INLINE void releaseProxy(ObjectProxy* proxy) {

        // Mark this proxy as empty. Reuse necessarily-non-null-for-in-use-proxies ptr as flag
        proxy->cell = 0;

        // Align on 64 boundary to get line
        uint32_t lineAddr = reinterpret_cast<uint32_t>(proxy) & 0xFFFFFFC0;

        // get proxy line
        ProxyLine* proxyLine = reinterpret_cast<ProxyLine*>(lineAddr);

        // Check if line is ready to be removed: when the 2 objects are empty
        if ((reinterpret_cast<uint32_t>(proxyLine->proxy[0].cell) | reinterpret_cast<uint32_t>(proxyLine->proxy[1].cell))==0) {
            // TODO: freeProxy becomes a double linked list => remove this line from the freeProxy list
            releaseLine(proxyLine);
        }
        // else TODO: add in double-linked list to increase memory re-use and decrease fragmentation

    }

};

#endif
