/*
    Neighand: Neighborhood Handling library

    The goal of this project is to find 3D neighbors efficiently.
    Read the documentation contained in the article "Spherical
    Indexing for Neighborhood Queries" for full details.

    This file defines the data structures that are used for this
    project.

    Nicolas Brodu, 2006
    Code released according to the GNU GPL, v2 or above.
*/
#ifndef NEIGHAND_STRUCTURES_HPP
#define NEIGHAND_STRUCTURES_HPP

// concat the arguments
#define COMPILE_TIME_ASSERT_CONCAT_HELPER(X,Y) X ## Y
// expand the macro arguments, including __LINE__
#define COMPILE_TIME_ASSERT_CONCAT(X,Y) COMPILE_TIME_ASSERT_CONCAT_HELPER(X,Y)
template<int line, bool x> struct AssertionFailedLine;
template<int line> struct AssertionFailedLine<line, true> {};
#define COMPILE_TIME_ASSERT(x) \
struct COMPILE_TIME_ASSERT_CONCAT(CompileTimeAssertionFailed_Line,__LINE__) : public AssertionFailedLine<__LINE__,bool(x)> { };

// Assert a few assumptions
COMPILE_TIME_ASSERT(sizeof(FloatType)==4)
// sizeof(char) is always 1, but char is not always 8 bits... => reject these machines
COMPILE_TIME_ASSERT(sizeof(uint32_t)==4)
// We need memory pointers to be 4 bytes too
COMPILE_TIME_ASSERT(sizeof(void*)==4)
// We would need to check that float endianity is the same as integer endianity
// This apparently cannot be done at compile time according to current language specs
// Specific system flags could be used, like FLOAT_BYTE_ORDER and BYTE_ORDER macros.

struct CellEntry;

// What is passed to the DB user
struct ObjectProxy {
    FloatType x;                        // align x/y/z on 16-bytes boundary, plan for SSE2 extension
    FloatType y;                        //
    FloatType z;                        //
    ObjectProxy* next;                  // next proxy in the cell list
    void* object;                       // user object
    ObjectProxy* prev;                  // prev proxy in the cell list
    uint32_t cellIndex;                 // index of the top-level cell
    CellEntry* cell;
};
COMPILE_TIME_ASSERT(sizeof(ObjectProxy)==32)

// Main object for the array of cells. Also maintains the non-empty lists
// DESIGN CHOICE:
// - This cell size is limited to 8 because array addressing *8 is builtin on x86, *16 is not
// - The linked list was initially double-linked, which forces cell size 16
//   It is now single-linked, and ordered in increasing cell indices.
//   => the memory access is much faster when running through the non-empty list
struct CellEntry {
    ObjectProxy* objects;      // 0 if empty cell
    CellEntry* nextNonEmpty;   // single-linked list
};
COMPILE_TIME_ASSERT(sizeof(CellEntry)==8)

struct ClosestNeighbor {
    void* object;
    FloatType squaredDistance;
};

// Basic linked list of user objects. Please use the NeighborListHolder below.
struct NeighborList {
    void* object;
    NeighborList* next;
};

struct LineManager;

// Handles a NeighborList returned by a neighbor search.
// Owns its neighbors list field.
// Copy is a deep copy that creates a new list.
// Does not own its manager field => lifetime is that of the main neighbor handler object
struct NeighborListHolder {
    uint32_t size;                // Total number of neighbors
    NeighborList* neighbors;      // Linked list
    NeighborListHolder() : size(0), neighbors(0), lineManager(0) {}

    NeighborListHolder(const NeighborListHolder&);
    NeighborListHolder& operator=(const NeighborListHolder&);
    ~NeighborListHolder();

// Exposed for template friend messiness reasons, should be protected
    LineManager* lineManager;
    void destroy();
};


#endif
