/*
    Neighand: Neighborhood Handling library

    The goal of this project is to find 3D neighbors efficiently.
    Read the documentation contained in the article "Spherical
    Indexing for Neighborhood Queries" for full details.

    This file defines a helper object that is used when the region
    of interest is cyclic in X and Y, but not in Z.

    Nicolas Brodu, 2006
    Code released according to the GNU GPL, v2 or above.
*/

// Specialization: wrap in X/Y, but do not wrap in Z
// This is especially handy for making cyclic terrains, etc.
//
// Packed index structure
// |0..0|E|idx|
// With:
// - E: external bit
// - idx: . either the internal |z|y|x| pattern for inside cells
//        . or the external region number as 1|0|0|0|
// => This gives 2^(exp2divx+exp2divy+exp2divz)+1 entries in the cells array
// Actually, reserve one for a "dummy" external that is always empty for quick rejection => +2 is the true array size
//
// Unpacked index structure:
// |0..0|y|nz-1 times 0|bz|z|ny times 0|x|
// With:
// - bz: bit for z outside flag.
//   The trick is to use 2^(exp2+1) complement => this bit is set both for <0 and >=2^exp, exactly what we want
//   Since the offsets are added from a position inside the main region, there cannot be overflow

template <int exp2divx, int exp2divy, int exp2divz>
struct WrapHelper<exp2divx, exp2divy, exp2divz, true, true, false, false> {


// See general code comments in ffff


COMPILE_TIME_ASSERT(exp2divx*2+exp2divy*2+exp2divz<=29)

    enum {
        MaxDQ = ((1 << (exp2divx-1))-1) * ((1 << (exp2divx-1))-1)
              + ((1 << (exp2divy-1))-1) * ((1 << (exp2divy-1))-1)
              + ( (exp2divz>1) ? (((1 << exp2divz) - 2) * ((1 << exp2divz) - 2)) : 0 ),
        MaxVolume = (1<<exp2divx)*(1<<exp2divy)*(1<<exp2divz),
        ArraySize = (1 << (exp2divx+exp2divy+exp2divz)) + 2,
        OutsideIndex = 1 << (exp2divx + exp2divy + exp2divz),
        OutsideIndexReal = (1 << (exp2divx + exp2divy + exp2divz))+1,
        WMaskX = ((1 << exp2divx) - 1),
        WMaskY = ((1 << exp2divy) - 1) << exp2divx,
        ShiftY = exp2divx*2 + exp2divy + exp2divz,
        ShiftZ = exp2divx + exp2divy,
        CorrConst = (1 << (exp2divx + exp2divy + exp2divz)) | (1 << exp2divx),
        MaxEdges = ((1 << exp2divx) - 1) | (((1 << exp2divy) - 1) << (exp2divx*2+exp2divy+exp2divz)) | (((1 << exp2divz) - 1) << (exp2divx+exp2divy)),
        UnpackedXMask = (1 << exp2divx) - 1,
        UnpackedYMask = ((1 << exp2divy) - 1) << (exp2divx*2+exp2divy+exp2divz),
        UnpackedZMask = ((1 << exp2divz) - 1) << (exp2divx+exp2divy),
        XMask = (1 << exp2divx) - 1,
        YMask = ((1 << exp2divy) - 1) << exp2divx,
        ZMask = ((1 << exp2divz) - 1) << (exp2divx+exp2divy),
        UnpackMask = ((1 << exp2divy) - 1) << exp2divx,
        UnpackMaskInv = ~(((1 << exp2divy) - 1) << exp2divx),
        ShiftInsert = exp2divx + exp2divy + exp2divz,
        WrapMask = ((1 << exp2divx) - 1) | (((1 << exp2divy) - 1) << (exp2divx*2+exp2divy+exp2divz)) | (((1 << exp2divz) - 1) << (exp2divx+exp2divy)),
        OutPackMask = ((1 << exp2divz) << (exp2divx+exp2divy)),
        ZTo1OutShift = exp2divx+exp2divy+exp2divz,
        PackMask = (1 << (exp2divx + exp2divy + exp2divz)) - 1,
        EXMask = (1 << exp2divx) - 1,
        EYMask = (1 << exp2divy) - 1,
        EZMask = (1 << (exp2divz+1)) - 1
    };

    uint32_t outsideFlag;
    CellEntry* outsideCell;

    FloatType minx, miny, minz, maxx, maxy, maxz, cellSize, cellSizeInv, cellSizeSquared;

    WrapHelper(FloatType _minx, FloatType _miny, FloatType _minz, FloatType _cellSize) :
        outsideFlag(0),
        minx(_minx), miny(_miny), minz(_minz),
        maxx(_minx + _cellSize * (1<<exp2divx)),
        maxy(_miny + _cellSize * (1<<exp2divy)),
        maxz(_minz + _cellSize * (1<<exp2divz)),
        cellSize(_cellSize), cellSizeSquared(_cellSize*_cellSize)
    {
        int roundmode = fegetround();
        fesetround(FE_TOWARDZERO);
        cellSizeInv = FloatType(1.0f) / _cellSize;
        fesetround(roundmode);
    }

    NEIGHAND_INLINE uint_fast32_t getCellIndexForWorldPosition(FloatType dx, FloatType dy, FloatType dz) {

        dx = (dx - minx) * cellSizeInv;
        dy = (dy - miny) * cellSizeInv;

        FloatType maxTestz = maxz - dz;
        dz -= minz;

        // See comment in ffff version
        if ((
            (*reinterpret_cast<uint32_t*>(&dz)) | (*reinterpret_cast<uint32_t*>(&maxTestz) - 1)
        ) > 0x7FFFFFFF)
            return OutsideIndexReal;

        uint32_t idx = fastFloorInt(dx) & WMaskX;
        idx |= (fastFloorInt(dy) << exp2divx) & WMaskY;
        // z necessary inside now
        return idx | (uint32_t(dz * cellSizeInv) << ShiftZ);
    }

    NEIGHAND_INLINE uint_fast32_t getCellIndexForWorldPosition(FloatType x, FloatType y, FloatType z, int32_t& cellx, int32_t& celly, int32_t& cellz, FloatType &xcenter, FloatType &ycenter, FloatType &zcenter) NEIGHAND_REALLY_ALWAYS_INLINE {
        xcenter = (x - minx) * cellSizeInv;
        ycenter = (y - miny) * cellSizeInv;
        cellx = fastFloorInt(xcenter);
        celly = fastFloorInt(ycenter);

        FloatType maxTestz = maxz - z;
        z -= minz;

        zcenter = z * cellSizeInv;
        cellz = int32_t(zcenter);

        if ((
            (*reinterpret_cast<uint32_t*>(&z)) | (*reinterpret_cast<uint32_t*>(&maxTestz) - 1)
        ) > 0x7FFFFFFF)
            return OutsideIndexReal;

        return (cellx & WMaskX) | ((celly << exp2divx) & WMaskY) | (cellz << ShiftZ);
    }


    // wrap in X/Y, not Z
    NEIGHAND_INLINE FloatType squaredDistance(FloatType dx, FloatType dy, FloatType dz) {
        dx = fastExp2Rem<exp2divx>(dx * cellSizeInv);
        dy = fastExp2Rem<exp2divy>(dy * cellSizeInv);
        return (dx*dx+dy*dy) * cellSizeSquared + dz*dz;
    }

    NEIGHAND_INLINE void setOutsideCell(CellEntry* _outsideCell) {
        outsideCell = _outsideCell;
    }

    NEIGHAND_INLINE void clearOutsideFlag() {
        outsideFlag = 0;
    }

    // xyz are in world coordinates, inside, not in cell units
    NEIGHAND_INLINE void flagOutside(FloatType x, FloatType y, FloatType z, FloatType d) {

        // center+radius used to check if sphere intersects one of the 2 Z planes
        FloatType maxTestz = maxz - (z+d);
        z -= minz + d;

        outsideFlag = (outsideCell->objects!=0);
        outsideFlag &= ((*reinterpret_cast<uint32_t*>(&z)) | (*reinterpret_cast<uint32_t*>(&maxTestz) - 1)) >> 31;
    }

    NEIGHAND_INLINE bool dsqImpliesOutside(FloatType x, FloatType y, FloatType z, FloatType dsq) NEIGHAND_REALLY_ALWAYS_INLINE {

        if (!outsideCell->objects) return false;

        FloatType d = maxz - z; d *= d; d -= dsq;
        FloatType d2 = z - minz; d2 *= d2; d2 -= dsq;

        return (*reinterpret_cast<uint32_t*>(&d) | *reinterpret_cast<uint32_t*>(&d2)) >> 31;
    }

    NEIGHAND_INLINE void flagOutside() NEIGHAND_REALLY_ALWAYS_INLINE {
        outsideFlag = (outsideCell->objects!=0);
    }

    NEIGHAND_INLINE bool outsideIsFlagged() {
        return outsideFlag!=0;
    }

    NEIGHAND_INLINE bool outsideIsNonEmpty() {
        return outsideCell->objects != 0;
    }

    NEIGHAND_INLINE ObjectProxy* getExternalRegionObjectList() {
        return outsideCell->objects;
    }

    // center position in world coordinates, outside, return the parallelepiped cells
    // Unlike ffff case, there could be several "blocks" or sub-regions for the parallelepiped
    // on the different sides for x/y, if the parallelepiped is over the edge.
    // => choose to use negative (or >max) index so there is only one region, and then xyzToPackedIndex will handle the situation
    NEIGHAND_INLINE uint_fast32_t getIntersectParallelepiped(int32_t &mincellx, int32_t &mincelly, int32_t &mincellz, int32_t &maxcellx, int32_t &maxcelly, int32_t &maxcellz, FloatType cx, FloatType cy, FloatType cz, FloatType d) {

        // a,b,c cube englobing the query sphere
        FloatType mina = (cx - d - minx) * cellSizeInv;
        FloatType maxa = (cx + d - minx) * cellSizeInv;
        FloatType minb = (cy - d - miny) * cellSizeInv;
        FloatType maxb = (cy + d - miny) * cellSizeInv;
        FloatType minc = cz - d;
        FloatType maxc = cz + d;

        // x/y are directly mapped and MUST be negative if need be for looping around parallelepiped cells
        // xyzToPackedIndex will handle the situation later on
        mincellx = int32_t(mina) - uint32_t(*reinterpret_cast<uint32_t*>(&mina) >> 31);
        mincelly = int32_t(minb) - uint32_t(*reinterpret_cast<uint32_t*>(&minb) >> 31);
        maxcellx = int32_t(maxa) - uint32_t(*reinterpret_cast<uint32_t*>(&maxa) >> 31);
        maxcelly = int32_t(maxb) - uint32_t(*reinterpret_cast<uint32_t*>(&maxb) >> 31);

        // compare with main region bound
        FloatType c1 = maxz - minc;
        maxc -= minz;

        // Test if any one of the above main region bound is negative => sphere is completely outside
        // It's better to add one test here than have a null condition on y or z in the main for loops
        // => report the condition on x, so the outermost loop is not entered
        if (( (*reinterpret_cast<uint32_t*>(&maxc)) | (*reinterpret_cast<uint32_t*>(&c1) - 1)
        ) > 0x7FFFFFFF) {
            // ensure mincellx > maxcellx for the loop to fail
            mincellx = 1; maxcellx = 0;
            mincelly = 1; maxcelly = 0;
            mincellz = 1; maxcellz = 0;
            return 0;
        }

        // finish the convertion from world coordinates to cell space
        minc -= minz;

        // Force signed cast from float, then assigned to unsigned int
        mincellz = static_cast<int32_t>(minc * cellSizeInv);
        maxcellz = static_cast<int32_t>(maxc * cellSizeInv);

        // Intersect by saturating z min/max to the main region bound. See ffff code
        mincellz &= (uint32_t(mincellz) >> 31) - 1;
        maxcellz -= ((1 << exp2divz) - 1);
        maxcellz &= -(uint32_t(maxcellz) >> 31);
        maxcellz += (1 << exp2divz);

        // maxcell - mincell + 1 in each dim
        uint_fast32_t volume = (maxcellx - mincellx + 1) * (maxcelly - mincelly + 1) * (maxcellz - mincellz);

        maxcellz -= 1;

        return volume;
    }

    // Center is outside, get distance to main region
    NEIGHAND_INLINE FloatType getSquaredDistanceToMainRegion(FloatType, FloatType, FloatType cz) {
        // Compute distances to the 2 z-planes
        FloatType c1 = minz - cz;
        FloatType c2 = maxz - cz;

        c1 *= c1; c2 *= c2;

        // Get min of the 2
        c1 -= c2;
        // generate bitmask from sign => if >=0 then AND with 0, if negative then AND with 0xFFFFFFFF
        *reinterpret_cast<uint32_t*>(&c1) &= -((*reinterpret_cast<uint32_t*>(&c1)) >> 31);
        // c1 - c2 >= 0 => 0 => +c2 = c2, and c1 - c2 < 0 => c1 - c2 => +c2 = c1
        // never mind if there is loss of precision, worse case is processing the main region in the main algo
        c1 += c2;

        return c1;
    }


    // center position in world coordinates, inside, return the cells in a parallelepiped bounding the internal
    // portion of the sphere.
    // return the volume = total number of cells in the parallelepiped
    NEIGHAND_INLINE uint_fast32_t getInternalParallelepiped(int32_t &mincellx, int32_t &mincelly, int32_t &mincellz, int32_t &maxcellx, int32_t &maxcelly, int32_t &maxcellz, int32_t &mincellx_unbounded, int32_t &mincelly_unbounded, int32_t &mincellz_unbounded, int32_t &maxcellx_unbounded, int32_t &maxcelly_unbounded, int32_t &maxcellz_unbounded, FloatType x_cellSpace, FloatType y_cellSpace, FloatType z_cellSpace, FloatType d_cellSpace) NEIGHAND_REALLY_ALWAYS_INLINE {

        mincellx_unbounded = mincellx = fastFloorInt(x_cellSpace - d_cellSpace);
        maxcellx_unbounded = maxcellx = fastFloorInt(x_cellSpace + d_cellSpace);
        if (d_cellSpace >= FloatType(1<<(exp2divx-1)) - FloatType(0.5f)) {
            maxcellx = (1<<exp2divx) - 1;
            mincellx = 0;
        }
        uint_fast32_t volume = maxcellx - mincellx + 1;

        mincelly_unbounded = mincelly = fastFloorInt(y_cellSpace - d_cellSpace);
        maxcelly_unbounded = maxcelly = fastFloorInt(y_cellSpace + d_cellSpace);
        if (d_cellSpace >= FloatType(1<<(exp2divy-1)) - FloatType(0.5f)) {
            maxcelly = (1<<exp2divy) - 1;
            mincelly = 0;
        }

        volume *= maxcelly - mincelly + 1;

        mincellz = fastFloorInt(z_cellSpace - d_cellSpace);
        mincellz_unbounded = mincellz;
        maxcellz = fastFloorInt(z_cellSpace + d_cellSpace);
        maxcellz_unbounded = maxcellz;
        mincellz &= (uint32_t(mincellz)>>31) -1;
        maxcellz -= (1 << exp2divz) - 1;
        maxcellz &= -(uint32_t(maxcellz)>>31);
        maxcellz += (1 << exp2divz);
        volume *= maxcellz - mincellz;
        maxcellz -= 1;

        return volume;
    }



    NEIGHAND_INLINE uint_fast32_t getInternalParallelepipedAndFlagOutside(int32_t &mincellx, int32_t &mincelly, int32_t &mincellz, int32_t &maxcellx, int32_t &maxcelly, int32_t &maxcellz, int32_t &mincellx_unbounded, int32_t &mincelly_unbounded, int32_t &mincellz_unbounded, int32_t &maxcellx_unbounded, int32_t &maxcelly_unbounded, int32_t &maxcellz_unbounded, FloatType x_cellSpace, FloatType y_cellSpace, FloatType z_cellSpace, FloatType d_cellSpace) NEIGHAND_REALLY_ALWAYS_INLINE {

        mincellx_unbounded = mincellx = fastFloorInt(x_cellSpace - d_cellSpace);
        maxcellx_unbounded = maxcellx = fastFloorInt(x_cellSpace + d_cellSpace);
        if (d_cellSpace >= FloatType(1<<(exp2divx-1)) - FloatType(0.5f)) {
            maxcellx = (1<<exp2divx) - 1;
            mincellx = 0;
        }
        uint_fast32_t volume = maxcellx - mincellx + 1;

        mincelly_unbounded = mincelly = fastFloorInt(y_cellSpace - d_cellSpace);
        maxcelly_unbounded = maxcelly = fastFloorInt(y_cellSpace + d_cellSpace);
        if (d_cellSpace >= FloatType(1<<(exp2divy-1)) - FloatType(0.5f)) {
            maxcelly = (1<<exp2divy) - 1;
            mincelly = 0;
        }

        volume *= maxcelly - mincelly + 1;

        maxcellz = fastFloorInt(z_cellSpace + d_cellSpace);
        z_cellSpace -= d_cellSpace;
        maxcellz_unbounded = maxcellz;
        mincellz = fastFloorInt(z_cellSpace);
        mincellz_unbounded = mincellz;
        maxcellz -= (1 << exp2divz) - 1;
        mincellz &= (uint32_t(mincellz)>>31) -1;

        // see comments in ffff
        outsideFlag = (outsideCell->objects!=0);
        outsideFlag &= uint32_t( (*reinterpret_cast<uint32_t*>(&z_cellSpace)) | (~maxcellz)) >> 31;

        maxcellz &= -(uint32_t(maxcellz)>>31);
        maxcellz += (1 << exp2divz);

        // maxcell - mincell + 1 in each dim
        volume *= maxcellz - mincellz;

        maxcellz -= 1;

        return volume;
    }


// Static functions below

    static NEIGHAND_INLINE bool isInside(uint_fast32_t packedIndex) NEIGHAND_REALLY_ALWAYS_INLINE {
        return packedIndex < OutsideIndex;
    }

    static NEIGHAND_INLINE int isUnpackedOutside(uint_fast32_t unpackedIndex) NEIGHAND_REALLY_ALWAYS_INLINE {
        return unpackedIndex & OutPackMask;
    }

    static NEIGHAND_INLINE uint_fast32_t unpack(uint_fast32_t idx) NEIGHAND_REALLY_ALWAYS_INLINE {
        return ((idx & UnpackMask) << ShiftInsert) | (idx & UnpackMaskInv);
    }

    // This function may not return the real outside cell, only the dummy index one (always empty)
    static NEIGHAND_INLINE uint_fast32_t packwrap(uint_fast32_t idx) NEIGHAND_REALLY_ALWAYS_INLINE {
        idx &= WrapMask;
        return ((idx >> ShiftInsert) | idx) & PackMask;
    }

    // This function does NOT always receive valid arguments in the ttff case, masking is necessary for x/y, z is OK
    static NEIGHAND_INLINE uint_fast32_t xyzToPackedIndex(int_fast32_t cellx, int_fast32_t celly, int_fast32_t cellz) NEIGHAND_REALLY_ALWAYS_INLINE {
        return (cellx & EXMask) | ((celly & EYMask) << exp2divx) | (cellz << ShiftZ);
    }

    static NEIGHAND_INLINE void packedIndexToXYZ(uint32_t packedCellIndex, int32_t& cellx, int32_t& celly, int32_t& cellz) NEIGHAND_REALLY_ALWAYS_INLINE {
        cellx = packedCellIndex & XMask;
        celly = (packedCellIndex & YMask) >> exp2divx;
        cellz = (packedCellIndex & ZMask) >> ShiftZ;
    }

    static NEIGHAND_INLINE int32_t extractX(uint32_t offset) NEIGHAND_REALLY_ALWAYS_INLINE {
        int32_t ret = offset & EXMask;
        ret |= -(ret & (1<< (exp2divx-1))); // extend bit sign
        return ret;
    }
    static NEIGHAND_INLINE int32_t extractY(uint32_t offset) NEIGHAND_REALLY_ALWAYS_INLINE {
        int32_t ret = (offset >> ShiftY) & EYMask;
        ret |= -(ret & (1<< (exp2divy-1))); // extend bit sign
        return ret;
    }
    static NEIGHAND_INLINE int32_t extractZ(uint32_t offset) NEIGHAND_REALLY_ALWAYS_INLINE {
        int32_t ret = (offset >> ShiftZ) & EZMask;
        ret |= -(ret & (1<< exp2divz)); // extend bit sign
        return ret;
    }

    static NEIGHAND_INLINE FloatType wrapCellDeltaX(FloatType delta) NEIGHAND_REALLY_ALWAYS_INLINE {
        return fastExp2Rem<exp2divx>(delta);
    }
    static NEIGHAND_INLINE FloatType wrapCellDeltaY(FloatType delta) NEIGHAND_REALLY_ALWAYS_INLINE {
        return fastExp2Rem<exp2divy>(delta);
    }
    static NEIGHAND_INLINE FloatType wrapCellDeltaZ(FloatType delta) NEIGHAND_REALLY_ALWAYS_INLINE {
        return delta;
    }

    static NEIGHAND_INLINE FloatType distCompHeuristicConstant() NEIGHAND_REALLY_ALWAYS_INLINE {
        return FloatType(1.00001f);
    }



    // take two cells at unpacked index idx1 and idx2, both inside
    // return the minimum squared distance any point in these cells can have, in cell units
    static NEIGHAND_INLINE uint_fast32_t getWrappedDeltaUnpackedIndexDQ(uint_fast32_t idx1, uint_fast32_t idx2) NEIGHAND_REALLY_ALWAYS_INLINE {
        // 1. Extract X/Y/Z
        uint32_t dx = (idx1 & UnpackedXMask) - (idx2 & UnpackedXMask);
        uint32_t dy = ((idx1 & UnpackedYMask) >> ShiftY) - ((idx2 & UnpackedYMask) >> ShiftY);
        uint32_t dz = ((idx1 & UnpackedZMask) >> ShiftZ) - ((idx2 & UnpackedZMask) >> ShiftZ);
        // 2. bit signs, on exp2 bits for X/Y
        uint32_t bsx = (dx >> (exp2divx-1)) & 1;
        uint32_t bsy = (dy >> (exp2divy-1)) & 1;
        uint32_t bsz = dz >> 31;
        // 3. compute abs(xyz) - 1, expand on 32 bits for X/Y
        dx = (((dx ^ (-bsx)) + bsx) & ((1 << exp2divx)-1) ) - 1;
        dy = (((dy ^ (-bsy)) + bsy) & ((1 << exp2divy)-1) ) - 1;
        dz = (dz ^ (-bsz)) + bsz - 1;
        // compute (|xyz|-1)^2 if |xyz|>1, or 0
        // Note: 0xFFFFFFFF squared is 1 even in unsigned arithmetic
        //       (2^32-1)*(2^32-1) = 2^64 + 1 - 2*2^32 => only 1 remains on a 32 bits variable
        dx = dx * dx - (dx >> 31);
        dy = dy * dy - (dy >> 31);
        dz = dz * dz - (dz >> 31);
        // return min dist squared from all 64 combinations of cube summits
        return dx+dy+dz;
    }

    enum {
        ShaveXN = 1,
        ShaveXP = 2,
        ShaveYN = 4,
        ShaveYP = 8,
        ShaveZN = 16,
        ShaveZP = 32
    };

    // The sphere offsets pre-computation depends on wrapping
    static NEIGHAND_INLINE void buildDistanceArray(uint32_t* &baseDistanceArray, uint32_t* &sphereOffsets, uint32_t** shavingComplement, uint32_t** shavingOffsets, uint32_t& maxWorldDist32, FloatType* &distanceVolumeTable, const char* initFileName) NEIGHAND_REALLY_ALWAYS_INLINE {

        // N - version/exp2divx/y/z on 4 bits each = 16 bits - ttff in binary = 1100
        uint32_t version = 1; // binary file format version, not software release version
        uint32_t magic = 'N' | ( ((version<<12)|((exp2divx<<8)|(exp2divy<<4)|exp2divz))<< 8) | ( 0xC << 24);

        // try loading the init file first
        std::ifstream initFile;
        initFile.open(initFileName, std::ios::binary);
        if (initFile.good()) {

            uint32_t counts;
            // read a magic marker: don't allocate mem otherwise
            initFile.read(reinterpret_cast<char*>(&counts), sizeof(uint32_t));
            if (counts == magic) {

                initFile.read(reinterpret_cast<char*>(&maxWorldDist32), sizeof(uint32_t));
                distanceVolumeTable = new FloatType[maxWorldDist32+1];
                initFile.read(reinterpret_cast<char*>(distanceVolumeTable), (maxWorldDist32+1)*sizeof(FloatType));

                void* newMem;
                initFile.read(reinterpret_cast<char*>(&counts), sizeof(uint32_t));
                posix_memalign(&newMem, 64, counts*sizeof(uint32_t));
                sphereOffsets = reinterpret_cast<uint32_t*>(newMem);
                initFile.read(reinterpret_cast<char*>(sphereOffsets), counts*sizeof(uint32_t));

                posix_memalign(&newMem, 64, (MaxDQ+2)*sizeof(uint32_t));
                baseDistanceArray = reinterpret_cast<uint32_t*>(newMem);
                initFile.read(reinterpret_cast<char*>(baseDistanceArray), (MaxDQ+2)*sizeof(uint32_t));
                ++baseDistanceArray;

                for (int shaving=0; shaving<64; ++shaving) {
                    initFile.read(reinterpret_cast<char*>(&counts), sizeof(uint32_t));
                    posix_memalign(&newMem, 64, counts*sizeof(uint32_t));
                    shavingOffsets[shaving] = reinterpret_cast<uint32_t*>(newMem);
                    initFile.read(reinterpret_cast<char*>(shavingOffsets[shaving]), counts*sizeof(uint32_t));

                    posix_memalign(&newMem, 64, (MaxDQ+2)*sizeof(uint32_t));
                    shavingComplement[shaving] = reinterpret_cast<uint32_t*>(newMem);
                    initFile.read(reinterpret_cast<char*>(shavingComplement[shaving]), (MaxDQ+2)*sizeof(uint32_t));
                    ++shavingComplement[shaving];
                }
                bool success = initFile.good();
                // read a magic marker again for end of file
                initFile.read(reinterpret_cast<char*>(&counts), sizeof(uint32_t));
                success &= (counts == magic);
                initFile.close();
                if (success) return;

                // Error! deallocate memory and recompute the tables
                delete [] distanceVolumeTable;
                for (int shaving = 0; shaving < 64; ++shaving) {
                    free(shavingComplement[shaving]-1);
                    free(shavingOffsets[shaving]);
                }
                free(baseDistanceArray-1);
                free(sphereOffsets);
            }
        }

        std::vector<uint32_t>* baseOffsets = new std::vector<uint32_t>[MaxDQ+2];
        ++baseOffsets; // so index -1 is valid
        uint32_t counts = 0;
        std::vector<uint32_t>* sOffsets[64];
        uint32_t sCounts[64];
        for (int i=0; i<64; ++i) {
            sOffsets[i] = new std::vector<uint32_t>[MaxDQ+1]; sCounts[i]=0;
        }

        maxWorldDist32 = uint32_t(sqrtf(MaxDQ)*32.f+55.425626f);
        distanceVolumeTable = new FloatType[maxWorldDist32+1];
        for (uint32_t i=0; i<=maxWorldDist32; ++i) distanceVolumeTable[i] = FloatType(0.f);

        // no wrap case in all 3 dims, offset from (0,0,0) is considered in either one of 8 directions, see below
        for (int z = 0; z<(1 << exp2divz); ++z) {
            for (int y = 0; y<(1 << exp2divy); ++y) {
                for (int x = 0; x<(1 << exp2divx); ++x) {

                    // Offset only for the positive direction, others will be obtained by symmetry
                    uint32_t unpackedIndex = uint32_t(x) | (uint32_t(y) << ShiftY) | (uint32_t(z) << ShiftZ);

                    // Match the 8 summits of the first cube with the 8 of the second
                    // retain min dist
                    uint32_t dq = getWrappedDeltaUnpackedIndexDQ(0,unpackedIndex);

                    int32_t dbase = uint32_t(sqrtf(dq));

                    // extend bit signs
                    int32_t ex = x | -(x & (1<<(exp2divx-1)));
                    int32_t ey = y | -(y & (1<<(exp2divy-1)));
                    FloatType dx = (ex<0) ? (ex+1) : ex;
                    FloatType dy = (ey<0) ? (ey+1) : ey;

                    // handle the 2 z directions, the distances above are preserved by symmetry.
                    int zval[] = {z, -z};
                    // should not duplicate the cells at 0 offset
                    for (int iz = 0; iz <= (z!=0); ++iz) {
                        // Extended x/y/z, possibly negative
                        int ez = zval[iz];
                        FloatType dz = (ez<0) ? (ez+1) : ez;

                        // for each cell, subdivide center in 1000 cubes
                        for (int i=0; i<10; ++i) for (int j=0; j<10; ++j) for (int k=0; k<10; ++k) {
                            FloatType cx = i * 0.1f + 0.05f;
                            FloatType cy = j * 0.1f + 0.05f;
                            FloatType cz = k * 0.1f + 0.05f;

                            // distance to target cell.
                            FloatType d = 0.f;
                            if (ex!=0) {
                                FloatType tmp = dx - cx;
                                d += tmp * tmp;
                            }
                            if (ey!=0) {
                                FloatType tmp = dy - cy;
                                d += tmp * tmp;
                            }
                            if (ez!=0) {
                                FloatType tmp = dz - cz;
                                d += tmp * tmp;
                            }
                            // increment count
                            uint32_t dbin = uint32_t(sqrtf(d) * 32.f);
                            if (dbin>maxWorldDist32) dbin = maxWorldDist32;
                            distanceVolumeTable[dbin] += FloatType(1.0f);
                        }

                        if ((dq==0) && (x==0) && (y==0) && (z==0)) {
                            baseOffsets[-1].push_back(0);
                            ++counts;
                            continue;
                        }

                        uint32_t currentCellOffset = (uint32_t(ex) & EXMask)
                            | ((uint32_t(ey) & EYMask) << ShiftY)
                            | ((uint32_t(ez) & EZMask) << ShiftZ);

                        if ((ex!=0) && (ey!=0)) currentCellOffset |= 0x60000000;
                        if ((ex!=0) && (ez!=0)) currentCellOffset |= 0xA0000000;
                        if ((ey!=0) && (ez!=0)) currentCellOffset |= 0xC0000000;
                        if (x == (1<<(exp2divx-1))) currentCellOffset &= ~0x20000000;
                        if (y == (1<<(exp2divy-1))) currentCellOffset &= ~0x40000000;

                        baseOffsets[dq].push_back(currentCellOffset);
                        ++counts;

                        for (uint_fast32_t shaving = 0; shaving < 64; ++shaving) {

                            // shave off some offsets
                            if ((dq==0) && ((ex!=0) || (ey!=0) || (ez!=0))) {
                                if (
                                ((((ex<0) && (shaving&ShaveXN)) || ((ex>0) && (shaving&ShaveXP))) && (x != (1<<(exp2divx-1))))
                                || ((((ey<0) && (shaving&ShaveYN)) || ((ey>0) && (shaving&ShaveYP))) && (y != (1<<(exp2divy-1))))
                                || ((ez<0) && (shaving&ShaveZN)) || ((ez>0) && (shaving&ShaveZP))
                                ) {
                                    continue;
                                }
                            } else if (dq>0) {
                                if ((((ex<0) && (shaving&ShaveXN)) || ((ex>0) && (shaving&ShaveXP)) || (ex==0) || (ex==-1) || (ex==1))
                                && (((ey<0) && (shaving&ShaveYN)) || ((ey>0) && (shaving&ShaveYP)) || (ey==0) || (ey==-1) || (ey==1))
                                && (((ez<0) && (shaving&ShaveZN)) || ((ez>0) && (shaving&ShaveZP)) || (ez==0) || (ez==-1) || (ez==1))
                                && (x != (1<<(exp2divx-1)))
                                && (y != (1<<(exp2divy-1)))
                                ) {
                                    continue;
                                }
                            }

                            if ((dq>1) && (
                            ((((ex<=-dbase) && (shaving&ShaveXN)) || ((ex>=dbase) && (shaving&ShaveXP))) && (x != (1<<(exp2divx-1))))
                            || ((((ey<=-dbase) && (shaving&ShaveYN)) || ((ey>=dbase) && (shaving&ShaveYP))) && (y != (1<<(exp2divy-1))))
                            || ((ez<=-dbase) && (shaving&ShaveZN)) || ((ez>=dbase) && (shaving&ShaveZP))
                            )) {
                                continue;
                            }

                            sOffsets[shaving][dq].push_back(currentCellOffset);
                            ++sCounts[shaving];
                        }

                    } // end z switcher

                }
            }
        }

        // finish volume distance table computation
        // cumulate to get volume < dist
        for (uint32_t i=1; i<=maxWorldDist32; ++i)
            distanceVolumeTable[i] += distanceVolumeTable[i-1];
        // finally convert sum of elementary cells to weighted volume
        for (uint32_t i=0; i<=maxWorldDist32; ++i)
            distanceVolumeTable[i] = distanceVolumeTable[i] / FloatType(1000.f);

        // Build the base array
        void* newMem;
        posix_memalign(&newMem, 64, counts*sizeof(uint32_t));
        sphereOffsets = reinterpret_cast<uint32_t*>(newMem);
        posix_memalign(&newMem, 64, (MaxDQ+2)*sizeof(uint32_t));
        baseDistanceArray = reinterpret_cast<uint32_t*>(newMem) + 1;
        baseDistanceArray[-1] = -1;
        uint32_t ntotal = 0;

        // linearize the array of vectors, keep the entry positions
        for (int dq = -1; dq <= MaxDQ; ++dq) {
            for (unsigned int i = 0; i<baseOffsets[dq].size(); ++i) {
                sphereOffsets[ntotal++] = baseOffsets[dq][i];
            }
            baseDistanceArray[dq] = ntotal;
        }
        delete [] (baseOffsets - 1);

        // now build the shaving complement arrays
        for (int shaving=0; shaving<64; ++shaving) {
            posix_memalign(&newMem, 64, sCounts[shaving]*sizeof(uint32_t));
            shavingOffsets[shaving] = reinterpret_cast<uint32_t*>(newMem);
            posix_memalign(&newMem, 64, (MaxDQ+2)*sizeof(uint32_t));
            shavingComplement[shaving] = reinterpret_cast<uint32_t*>(newMem);
            shavingComplement[shaving][0] = 0; // -1: used as init pos, no dup
            ++shavingComplement[shaving];

            uint32_t sTotal = 0;
            for (int dq = 0; dq <= MaxDQ; ++dq) {
                for (unsigned int i = 0; i<sOffsets[shaving][dq].size(); ++i) {
                    shavingOffsets[shaving][sTotal++] = sOffsets[shaving][dq][i];
                }
                shavingComplement[shaving][dq] = sTotal;
            }
            delete [] sOffsets[shaving];
        }

        // Done. Try saving the file
        std::ofstream outInitFile;
        outInitFile.open(initFileName, std::ios::binary);
        if (outInitFile.good()) {
            // write a magic marker
            outInitFile.write(reinterpret_cast<char*>(&magic), sizeof(uint32_t));
            outInitFile.write(reinterpret_cast<char*>(&maxWorldDist32), sizeof(uint32_t));
            outInitFile.write(reinterpret_cast<char*>(distanceVolumeTable), (maxWorldDist32+1)*sizeof(FloatType));
            outInitFile.write(reinterpret_cast<char*>(&counts), sizeof(uint32_t));
            outInitFile.write(reinterpret_cast<char*>(sphereOffsets), counts*sizeof(uint32_t));
            outInitFile.write(reinterpret_cast<char*>(baseDistanceArray-1), (MaxDQ+2)*sizeof(uint32_t));
            for (int shaving=0; shaving<64; ++shaving) {
                outInitFile.write(reinterpret_cast<char*>(&sCounts[shaving]), sizeof(uint32_t));
                outInitFile.write(reinterpret_cast<char*>(shavingOffsets[shaving]), sCounts[shaving]*sizeof(uint32_t));
                outInitFile.write(reinterpret_cast<char*>(shavingComplement[shaving]-1), (MaxDQ+2)*sizeof(uint32_t));
            }
            // write a magic marker again for end of file
            outInitFile.write(reinterpret_cast<char*>(&magic), sizeof(uint32_t));
            outInitFile.close();
        }
    }


};
