#include <sys/time.h>
#include <time.h>
#include <math.h>

#include <iostream>
#include <fstream>
using namespace std;

//#define NEIGHAND_SELECT_METHOD_STAT
#include "neighand.h"
using namespace neighand;

#include "lq.h"

// set these primary values

#ifndef BASE_EXP2
#warning Please define BASE_EXP2, WRAP_ARGUMENTS, and LOAD_RATIO
#define BASE_EXP2 4
#define WRAP_ARGUMENTS true,true,true
#define LOAD_RATIO 10.0f
#endif

// derived values

const FloatType loadRatio = LOAD_RATIO;
const int worldSizeInt = 1 << BASE_EXP2;
const FloatType worldSize = worldSizeInt;
const int Nagents = int(loadRatio * worldSize * worldSize * worldSize);
const FloatType cellSize = 1.234f * worldSize / 16.f;
#define TEMPLATE_SIZE BASE_EXP2,BASE_EXP2,BASE_EXP2
#define NEIGHAND_TEMPLATE_ARGUMENTS TEMPLATE_SIZE,WRAP_ARGUMENTS



struct Agent {
    int number;
    float x,y,z;
};

// The callback function for NH
// Do something that accesses the agent memory line
void callbackFunctionNH(void* object, void* userData) {
    Agent* agent = reinterpret_cast<Agent*>(object);
    ++agent->number;
    agent->number -= 1;
}

// The callback function for LQ
void callbackFunctionLQ(void* clientObject, float distanceSquared, void* clientQueryState) {
    Agent* agent = reinterpret_cast<Agent*>(clientObject);
    ++agent->number;
    agent->number -= 1;
}


FloatType getRate(timeval& start, timeval& stop, int reps = 1)
{
    long sec = stop.tv_sec - start.tv_sec;
    long usec = stop.tv_usec - start.tv_usec;
    if (usec<0) {
        sec -= 1;
        usec += 1000000L;
    }
    FloatType time = FloatType(sec) + FloatType(usec) * 1e-6f;
    return FloatType(reps) / time;
}

#define NEIGHAND_APPLY_XYZ_API 1
#define NEIGHAND_C_CALLBACK_API 1
#define NEIGHAND_CLOSEST_N_EQ_1 1

// A bin-lattice class using the neighand data structures & functor, for fair comparison
class BinLattice : public NeighborhoodHandler<NEIGHAND_TEMPLATE_ARGUMENTS> {
public:
    uint_fast32_t pvolume;

    BinLattice(FloatType minx, FloatType miny, FloatType minz, FloatType cellSize)
    : NeighborhoodHandler<NEIGHAND_TEMPLATE_ARGUMENTS>(minx, miny, minz, cellSize, "methodComp.bin") {}

    NEIGHAND_INLINE void applyToNeighbors(FloatType x, FloatType y, FloatType z, FloatType d, Callback f, void* userData) NEIGHAND_ALWAYS_INLINE {

        helper.clearOutsideFlag();

        FloatType dsq = d*d;
        FloatType d_cellSpace = d * helper.cellSizeInv;

        int32_t mincellx, mincelly, mincellz, maxcellx, maxcelly, maxcellz;
        int32_t cellx, celly, cellz;
        FloatType xcenter, ycenter, zcenter;

        int32_t mincellx_unbounded, mincelly_unbounded, mincellz_unbounded, maxcellx_unbounded, maxcelly_unbounded, maxcellz_unbounded;

        uint_fast32_t centerCellIndexPacked = helper.getCellIndexForWorldPosition(x,y,z, cellx, celly, cellz, xcenter, ycenter, zcenter);
        if (WrapHelper<NEIGHAND_TEMPLATE_ARGUMENTS>::isInside(centerCellIndexPacked)) {
            pvolume = helper.getInternalParallelepipedAndFlagOutside(mincellx, mincelly, mincellz, maxcellx, maxcelly, maxcellz, mincellx_unbounded, mincelly_unbounded, mincellz_unbounded, maxcellx_unbounded, maxcelly_unbounded, maxcellz_unbounded, xcenter, ycenter, zcenter, d_cellSpace);
        } else {
            pvolume = helper.getIntersectParallelepiped(mincellx, mincelly, mincellz, maxcellx, maxcelly, maxcellz, x, y, z, d);
            helper.flagOutside();
        }

        // Run through the external region list only if necessary
        if (helper.outsideIsFlagged()) {
            register ObjectProxy* plist = helper.getExternalRegionObjectList();
            #include "neighand_apply_processlines_cond.hpp"
        }

        // Total volume may happen in non-cyclic worlds where the sphere volume may exceed the region volume,
        // since offsets are counted in all directions
        // This cannot happen in all-cyclic worlds, since precisely the max "sphere" volume is bounded by the total volume
        if (pvolume==WrapHelper<NEIGHAND_TEMPLATE_ARGUMENTS>::MaxVolume) {
            // Faster than triple-loop
            ObjectProxy** cList = cachedLists;
            ObjectProxy** lastList = cachedLists + WrapHelper<NEIGHAND_TEMPLATE_ARGUMENTS>::MaxVolume;
            while (cList!=lastList) {
                register ObjectProxy* plist = *cList++;
                #define NEIGHAND_APPLY_CHECK_FIRST
                #include "neighand_apply_processlines_cond.hpp"
                #undef NEIGHAND_APPLY_CHECK_FIRST
            }
            return;
        }

        for (int_fast32_t cellz = mincellz; cellz <= maxcellz; ++cellz) {
            for (int_fast32_t celly = mincelly; celly <= maxcelly; ++celly) {
                for (int_fast32_t cellx = mincellx; cellx <= maxcellx; ++cellx) {
                    register ObjectProxy* plist = cachedLists[WrapHelper<NEIGHAND_TEMPLATE_ARGUMENTS>::xyzToPackedIndex(cellx,celly,cellz)];
                    #define NEIGHAND_APPLY_CHECK_FIRST
                    #include "neighand_apply_processlines_cond.hpp"
                    #undef NEIGHAND_APPLY_CHECK_FIRST
                }
            }
        }

        return;
    }

    NEIGHAND_INLINE int findClosestNeighbor(FloatType x, FloatType y, FloatType z, FloatType d, ClosestNeighbor* neighbor) NEIGHAND_ALWAYS_INLINE {

        helper.clearOutsideFlag();

        FloatType dsq = d*d;

        FloatType neighborSquaredDistance = dsq;
        void* neighborObject = 0;

        int32_t mincellx, mincelly, mincellz, maxcellx, maxcelly, maxcellz;
        int32_t cellx, celly, cellz;
        uint_fast32_t pvolume;
        FloatType xcenter, ycenter, zcenter;

        uint_fast32_t centerCellIndexPacked = helper.getCellIndexForWorldPosition(x,y,z, cellx, celly, cellz, xcenter, ycenter, zcenter);
        if (WrapHelper<NEIGHAND_TEMPLATE_ARGUMENTS>::isInside(centerCellIndexPacked)) {
            int32_t mincellx_unbounded, mincelly_unbounded, mincellz_unbounded, maxcellx_unbounded, maxcelly_unbounded,    maxcellz_unbounded;
            FloatType d_cellSpace = d * helper.cellSizeInv;
            pvolume = helper.getInternalParallelepipedAndFlagOutside(mincellx, mincelly, mincellz, maxcellx, maxcelly, maxcellz, mincellx_unbounded, mincelly_unbounded, mincellz_unbounded, maxcellx_unbounded, maxcelly_unbounded, maxcellz_unbounded, xcenter, ycenter, zcenter, d_cellSpace);
        } else {
            pvolume = helper.getIntersectParallelepiped(mincellx, mincelly, mincellz, maxcellx, maxcelly, maxcellz, x, y, z, d);
            helper.flagOutside();
        }

        // Run through the external region list only if necessary
        if (helper.outsideIsFlagged()) {
            register ObjectProxy* plist = helper.getExternalRegionObjectList();
            #include "neighand_closest_processlines.hpp"
        }

        if (pvolume==WrapHelper<NEIGHAND_TEMPLATE_ARGUMENTS>::MaxVolume) {
            // Faster than triple-loop
            ObjectProxy** cList = cachedLists;
            ObjectProxy** lastList = cachedLists + WrapHelper<NEIGHAND_TEMPLATE_ARGUMENTS>::MaxVolume;
            while (cList!=lastList) {
                register ObjectProxy* plist = *cList++;
                #define NEIGHAND_CLOSEST_CHECK_FIRST
                #include "neighand_closest_processlines.hpp"
                #undef NEIGHAND_CLOSEST_CHECK_FIRST
            }
        } else {
            for (int_fast32_t cellz = mincellz; cellz <= maxcellz; ++cellz) {
                for (int_fast32_t celly = mincelly; celly <= maxcelly; ++celly) {
                    for (int_fast32_t cellx = mincellx; cellx <= maxcellx; ++cellx) {
                        register ObjectProxy* plist = cachedLists[WrapHelper<NEIGHAND_TEMPLATE_ARGUMENTS>::xyzToPackedIndex(cellx,celly,cellz)];
                        #define NEIGHAND_CLOSEST_CHECK_FIRST
                        #include "neighand_closest_processlines.hpp"
                        #undef NEIGHAND_CLOSEST_CHECK_FIRST
                    }
                }
            }
        }

        neighbor->object = neighborObject;
        neighbor->squaredDistance = neighborSquaredDistance;
        return (neighbor[0].object != 0);
    }

};


// A class that only uses the spherical indexing algorithm (for the "inside" situation)
class SphereOnly : public NeighborhoodHandler<NEIGHAND_TEMPLATE_ARGUMENTS> {
public:
    uint32_t baseVolume;

    SphereOnly(FloatType minx, FloatType miny, FloatType minz, FloatType cellSize)
    : NeighborhoodHandler<NEIGHAND_TEMPLATE_ARGUMENTS>(minx, miny, minz, cellSize, "methodComp.bin") {}

    NEIGHAND_INLINE void applyToNeighbors(FloatType x, FloatType y, FloatType z, FloatType d, Callback f, void* userData) NEIGHAND_ALWAYS_INLINE {

        int32_t mincellx, mincelly, mincellz, maxcellx, maxcelly, maxcellz;
        int32_t mincellx_unbounded, mincelly_unbounded, mincellz_unbounded, maxcellx_unbounded, maxcelly_unbounded, maxcellz_unbounded;
        int32_t cellx, celly, cellz;
        FloatType xcenter, ycenter, zcenter;

        uint_fast32_t pvolume;

        FloatType dsq = d*d;
        helper.clearOutsideFlag();

        uint_fast32_t centerCellIndexPacked = helper.getCellIndexForWorldPosition(x,y,z, cellx, celly, cellz, xcenter, ycenter, zcenter);
        if (WrapHelper<NEIGHAND_TEMPLATE_ARGUMENTS>::isInside(centerCellIndexPacked)) {
            uint_fast32_t centerCellIndex = WrapHelper<NEIGHAND_TEMPLATE_ARGUMENTS>::unpack(centerCellIndexPacked);

            FloatType d_cellSpace = d * helper.cellSizeInv;
            pvolume = helper.getInternalParallelepipedAndFlagOutside(mincellx, mincelly, mincellz, maxcellx, maxcelly, maxcellz, mincellx_unbounded, mincelly_unbounded, mincellz_unbounded, maxcellx_unbounded, maxcelly_unbounded, maxcellz_unbounded, xcenter, ycenter, zcenter, d_cellSpace);

            uint32_t dmax = uint32_t(d_cellSpace);
            FloatType dsq_cellSpace = d_cellSpace*d_cellSpace;
            uint32_t lastdq = uint32_t(dsq_cellSpace);
            uint32_t dmax_p1 = dmax + 1;

            uint32_t shaving = ((cellx - mincellx_unbounded - dmax_p1) >> 31)
            | (((maxcellx_unbounded - cellx - dmax_p1) & 0x80000000) >> 30)
            | (((celly - mincelly_unbounded - dmax_p1) & 0x80000000) >> 29)
            | (((maxcelly_unbounded - celly - dmax_p1) & 0x80000000) >> 28)
            | (((cellz - mincellz_unbounded - dmax_p1) & 0x80000000) >> 27)
            | (((maxcellz_unbounded - cellz - dmax_p1) & 0x80000000) >> 26);

            int32_t dmaxSq_m1 = dmax*dmax - 1;

            FloatType dm1 = d_cellSpace - WrapHelper<NEIGHAND_TEMPLATE_ARGUMENTS>::distCompHeuristicConstant();
            int32_t dm1sq = uint32_t(dm1*dm1);
            if (dm1<FloatType(0.f)) dm1sq = -1;

            if (lastdq>MaxDQ) {
                lastdq = MaxDQ;
                dmaxSq_m1 = MaxDQ;
                dm1sq = MaxDQ;
                shaving = 0;
            }

            baseVolume = 0;
            uint32_t* baseMaxUp = &sphereOffsets[baseDistanceArray[dmaxSq_m1]];

            uint32_t* interOffsetList = &shavingOffsets[shaving][shavingComplement[shaving][dmaxSq_m1]];
            uint32_t* interMaxUp = &shavingOffsets[shaving][shavingComplement[shaving][lastdq]];

            xcenter -= cellx;
            ycenter -= celly;
            zcenter -= cellz;

            uint32_t dbldmax = dmax_p1 + dmax_p1 + 1;
            uint32_t dbldmax3 = dbldmax+dbldmax+dbldmax;
            uint32_t all_cache[dbldmax3];
            for (uint32_t i=0; i<dbldmax3; ++i) all_cache[i] = 0;
            // alias to mid-array to allow indexing by negative offsets
            FloatType* cachedDeltaX = reinterpret_cast<FloatType*>(all_cache) + dmax_p1;
            FloatType* cachedDeltaY = cachedDeltaX + dbldmax;
            FloatType* cachedDeltaZ = cachedDeltaY + dbldmax;

            uint32_t* baseOffsetList = &sphereOffsets[0];
            if (d_cellSpace > FloatType(1.73205081)) {
                d_cellSpace -= FloatType(1.73205081);
                baseOffsetList = &sphereOffsets[uint32_t(d_cellSpace * d_cellSpace)];
                register uint32_t* offsetlist = &sphereOffsets[0];
                while (offsetlist < baseOffsetList) {
                    uint32_t offset = *offsetlist++;
                    uint32_t unpackedCell = centerCellIndex + offset;
                    if (WrapHelper<NEIGHAND_TEMPLATE_ARGUMENTS>::isUnpackedOutside(unpackedCell)) continue;
                    register ObjectProxy* plist = cachedLists[WrapHelper<NEIGHAND_TEMPLATE_ARGUMENTS>::packwrap(unpackedCell)];
                    #define NEIGHAND_APPLY_CHECK_FIRST
                    #include "neighand_apply_processlines_uncond.hpp"
                    #undef NEIGHAND_APPLY_CHECK_FIRST
                }
            }

            register uint32_t* offsetlist = baseOffsetList;
            baseOffsetList = &sphereOffsets[baseDistanceArray[dm1sq]];
            while (offsetlist < baseOffsetList) {
                uint32_t offset = *offsetlist++;
                uint32_t unpackedCell = centerCellIndex + offset;
                if (WrapHelper<NEIGHAND_TEMPLATE_ARGUMENTS>::isUnpackedOutside(unpackedCell)) continue;
                register ObjectProxy* plist = cachedLists[WrapHelper<NEIGHAND_TEMPLATE_ARGUMENTS>::packwrap(unpackedCell)];
                #define NEIGHAND_APPLY_CHECK_FIRST
                #include "neighand_apply_processlines_cond.hpp"
                #undef NEIGHAND_APPLY_CHECK_FIRST
            }

            while (offsetlist < baseMaxUp) {
                uint32_t offset = *offsetlist++;
                uint32_t unpackedCell = centerCellIndex + offset;
                if (WrapHelper<NEIGHAND_TEMPLATE_ARGUMENTS>::isUnpackedOutside(unpackedCell)) continue;
                #include "neighand_apply_checkdist.hpp"
                register ObjectProxy* plist = cachedLists[WrapHelper<NEIGHAND_TEMPLATE_ARGUMENTS>::packwrap(unpackedCell)];
                #define NEIGHAND_APPLY_CHECK_FIRST
                #include "neighand_apply_processlines_cond.hpp"
                #undef NEIGHAND_APPLY_CHECK_FIRST
            }

            offsetlist = interOffsetList;
            while (offsetlist < interMaxUp) {
                uint32_t offset = *offsetlist++;
                uint32_t unpackedCell = centerCellIndex + offset;
                if (WrapHelper<NEIGHAND_TEMPLATE_ARGUMENTS>::isUnpackedOutside(unpackedCell)) continue;
                #include "neighand_apply_checkdist.hpp"
                register ObjectProxy* plist = cachedLists[WrapHelper<NEIGHAND_TEMPLATE_ARGUMENTS>::packwrap(unpackedCell)];
                #define NEIGHAND_APPLY_CHECK_FIRST
                #include "neighand_apply_processlines_cond.hpp"
                #undef NEIGHAND_APPLY_CHECK_FIRST
            }

            if (helper.outsideIsFlagged()) {
                register ObjectProxy* plist = helper.getExternalRegionObjectList();
                #include "neighand_apply_processlines_cond.hpp"
            }

            return;

        } else {
            pvolume = helper.getIntersectParallelepiped(mincellx, mincelly, mincellz, maxcellx, maxcelly, maxcellz, x, y, z, d);
            helper.flagOutside();
        }

        if (helper.outsideIsFlagged()) {
            register ObjectProxy* plist = helper.getExternalRegionObjectList();
            #include "neighand_apply_processlines_cond.hpp"
        }

        if (pvolume==WrapHelper<NEIGHAND_TEMPLATE_ARGUMENTS>::MaxVolume) {
            // Faster than triple-loop
            ObjectProxy** cList = cachedLists;
            ObjectProxy** lastList = cachedLists + WrapHelper<NEIGHAND_TEMPLATE_ARGUMENTS>::MaxVolume;
            while (cList!=lastList) {
                register ObjectProxy* plist = *cList++;
                #define NEIGHAND_APPLY_CHECK_FIRST
                #include "neighand_apply_processlines_cond.hpp"
                #undef NEIGHAND_APPLY_CHECK_FIRST
            }
            return;
        }

        for (int_fast32_t cellz = mincellz; cellz <= maxcellz; ++cellz) {
            for (int_fast32_t celly = mincelly; celly <= maxcelly; ++celly) {
                for (int_fast32_t cellx = mincellx; cellx <= maxcellx; ++cellx) {
                    register ObjectProxy* plist = cachedLists[WrapHelper<NEIGHAND_TEMPLATE_ARGUMENTS>::xyzToPackedIndex(cellx,celly,cellz)];
                    #define NEIGHAND_APPLY_CHECK_FIRST
                    #include "neighand_apply_processlines_cond.hpp"
                    #undef NEIGHAND_APPLY_CHECK_FIRST
                }
            }
        }

        return;

    }

};


// A class that only uses the non-empty list algorithm, for reference.
class NEList : public NeighborhoodHandler<NEIGHAND_TEMPLATE_ARGUMENTS> {
public:

    NEList(FloatType minx, FloatType miny, FloatType minz, FloatType cellSize)
    : NeighborhoodHandler<NEIGHAND_TEMPLATE_ARGUMENTS>(minx, miny, minz, cellSize, "methodComp.bin") {}

    NEIGHAND_INLINE void applyToNeighbors(FloatType x, FloatType y, FloatType z, FloatType d, Callback f, void* userData) NEIGHAND_ALWAYS_INLINE {

        FloatType dsq = d*d;

        for (CellEntry* entry = firstNonEmptyCell; entry!=0; entry = entry->nextNonEmpty) {
            register ObjectProxy* plist = entry->objects;
            #include "neighand_apply_processlines_cond.hpp"
        }

        if (helper.outsideIsNonEmpty()) {
            register ObjectProxy* plist = helper.getExternalRegionObjectList();
            #include "neighand_apply_processlines_cond.hpp"
        }

        return;
    }

};


int main() {

    cout << "Initializing" << endl;

    // Set the revert distances & sphere bias to your convenience
    NeighborhoodHandler<NEIGHAND_TEMPLATE_ARGUMENTS> nhCombi(10.f, 10.f, 10.f, cellSize, 0.0f, 0.0f, 1.0f, "methodComp.bin");

    BinLattice nhBinLattice(10.f, 10.f, 10.f, cellSize);
    SphereOnly nhSphere(10.f, 10.f, 10.f, cellSize);
    lqDB* db = lqCreateDatabase(10.f, 10.f, 10.f, worldSize*cellSize, worldSize*cellSize, worldSize*cellSize, worldSizeInt, worldSizeInt, worldSizeInt);

    Agent* agents = new Agent[Nagents];
    ObjectProxy** proxiesBinLattice = new ObjectProxy*[Nagents];
    ObjectProxy** proxiesSphere = new ObjectProxy*[Nagents];
    ObjectProxy** proxiesCombi = new ObjectProxy*[Nagents];
    lqClientProxy* proxiesLQ = new lqClientProxy[Nagents];

    // give the agents a unique number and a pre-computed position
    for (int i=0; i<Nagents; ++i) {
        agents[i].number = i+1;
        agents[i].x = 10.0f + worldSize*cellSize * (rand() / (RAND_MAX + 1.0f));
        agents[i].y = 10.0f + worldSize*cellSize * (rand() / (RAND_MAX + 1.0f));
        agents[i].z = 10.0f + worldSize*cellSize * (rand() / (RAND_MAX + 1.0f));
    }

    // Insertion
    for (int i=0; i<Nagents; ++i) {
        proxiesBinLattice[i] = nhBinLattice.insert(agents[i].x, agents[i].y, agents[i].z, &agents[i]);
        proxiesSphere[i] = nhSphere.insert(agents[i].x, agents[i].y, agents[i].z, &agents[i]);
        proxiesCombi[i] = nhCombi.insert(agents[i].x, agents[i].y, agents[i].z, &agents[i]);
        lqInitClientProxy(&proxiesLQ[i], &agents[i]);
        lqUpdateForNewLocation(db, &proxiesLQ[i], agents[i].x, agents[i].y, agents[i].z);
    }

    // precompute the center locations too so as to compare only
    // the distance and the algo type
    int Nqueries = 150;
    FloatType *cxpos = new FloatType[Nqueries];
    FloatType *cypos = new FloatType[Nqueries];
    FloatType *czpos = new FloatType[Nqueries];
    for (int i=0; i<Nqueries; ++i) {
        cxpos[i] = 10.0f + worldSize*cellSize * (rand() / (RAND_MAX + 1.0f));
        cypos[i] = 10.0f + worldSize*cellSize * (rand() / (RAND_MAX + 1.0f));
        czpos[i] = 10.0f + worldSize*cellSize * (rand() / (RAND_MAX + 1.0f));
    }

    ofstream closestValues("closestPerformances.txt");
    ofstream queryValues("queryPerformances.txt");

    // Distance loop: 0.1 increment in cellSize units, up to max dist
    ClosestNeighbor neighbor;
    timeval start, stop;
    for (int dinc = 0; dinc <= worldSizeInt*30/4; ++dinc) {
        FloatType distance = dinc / 10.f;
        cout << "Processing distance " << distance << " cell units " << flush;
        closestValues << distance << flush;
        queryValues << distance << flush;

        distance *= cellSize;

        // try to remove multitask jitter by eliminating some worse cases
        int N=40;

        // Find all objects within radius

        FloatType rate[40];
        FloatType avgrate = 0.f;

        for (int rep = 0; rep<N; ++rep) {
            gettimeofday(&start,0);
            for (int i=0; i<Nqueries; ++i) {
                nhBinLattice.findClosestNeighbor(cxpos[i], cypos[i], czpos[i], distance, &neighbor);
            }
            gettimeofday(&stop,0);
            // insert-sort the new rate
            FloatType reprate = getRate(start, stop, Nqueries);
            for (int j=0; j<rep; ++j) if (rate[j]<reprate) {
                FloatType tmp = rate[j]; rate[j] = reprate; reprate = tmp;
            }
            rate[rep] = reprate;
        }
        avgrate = 0.f;
        for (int i=0; i<30; ++i) avgrate += rate[i];
        avgrate /= 30.f;
        closestValues << " " << avgrate << flush;
        cout << "." << flush;

        for (int rep = 0; rep<N; ++rep) {
            gettimeofday(&start,0);
            for (int i=0; i<Nqueries; ++i) {
                lqFindNearestNeighborWithinRadius(db, cxpos[i], cypos[i], czpos[i], distance, 0);
            }
            gettimeofday(&stop,0);
            // insert-sort the new rate
            FloatType reprate = getRate(start, stop, Nqueries);
            for (int j=0; j<rep; ++j) if (rate[j]<reprate) {
                FloatType tmp = rate[j]; rate[j] = reprate; reprate = tmp;
            }
            rate[rep] = reprate;
        }
        avgrate = 0.f;
        for (int i=0; i<30; ++i) avgrate += rate[i];
        avgrate /= 30.f;
        closestValues << " " << avgrate << flush;
        cout << "." << flush;

        for (int rep = 0; rep<N; ++rep) {
            gettimeofday(&start,0);
            for (int i=0; i<Nqueries; ++i) {
                nhCombi.findClosestNeighbor(cxpos[i], cypos[i], czpos[i], distance, &neighbor);
            }
            gettimeofday(&stop,0);
            // insert-sort the new rate
            FloatType reprate = getRate(start, stop, Nqueries);
            for (int j=0; j<rep; ++j) if (rate[j]<reprate) {
                FloatType tmp = rate[j]; rate[j] = reprate; reprate = tmp;
            }
            rate[rep] = reprate;
        }
        avgrate = 0.f;
        for (int i=0; i<30; ++i) avgrate += rate[i];
        avgrate /= 30.f;
        closestValues << " " << avgrate << flush;
        cout << "." << flush;

        FloatType avgVol = 0.0f;
        for (int rep = 0; rep<N; ++rep) {
            gettimeofday(&start,0);
            for (int i=0; i<Nqueries; ++i) {
                nhSphere.applyToNeighbors(cxpos[i], cypos[i], czpos[i], distance, &callbackFunctionNH, 0);
#ifdef NEIGHAND_SELECT_METHOD_STAT
                avgVol += nhSphere.baseVolume;
#endif
            }
            gettimeofday(&stop,0);
            // insert-sort the new rate
            FloatType reprate = getRate(start, stop, Nqueries);
            for (int j=0; j<rep; ++j) if (rate[j]<reprate) {
                FloatType tmp = rate[j]; rate[j] = reprate; reprate = tmp;
            }
            rate[rep] = reprate;
        }
        avgrate = 0.f;
        for (int i=0; i<30; ++i) avgrate += rate[i];
        avgrate /= 30.f;
        queryValues << " " << avgrate << flush;
#ifdef NEIGHAND_SELECT_METHOD_STAT
        cout << ". s=" << (avgVol/(N*Nqueries)) << flush;
#else
        cout << "." << flush;
#endif
        avgVol = 0.0f;
        for (int rep = 0; rep<N; ++rep) {
            gettimeofday(&start,0);
            for (int i=0; i<Nqueries; ++i) {
                nhBinLattice.applyToNeighbors(cxpos[i], cypos[i], czpos[i], distance, &callbackFunctionNH, 0);
#ifdef NEIGHAND_SELECT_METHOD_STAT
                avgVol += nhBinLattice.pvolume;
#endif
            }
            gettimeofday(&stop,0);
            // insert-sort the new rate
            FloatType reprate = getRate(start, stop, Nqueries);
            for (int j=0; j<rep; ++j) if (rate[j]<reprate) {
                FloatType tmp = rate[j]; rate[j] = reprate; reprate = tmp;
            }
            rate[rep] = reprate;
        }
        avgrate = 0.f;
        for (int i=0; i<30; ++i) avgrate += rate[i];
        avgrate /= 30.f;
        queryValues << " " << avgrate << flush;
#ifdef NEIGHAND_SELECT_METHOD_STAT
        cout << ". v=" << (avgVol/(40*Nqueries)) << flush;
#else
        cout << "." << flush;
#endif

        for (int rep = 0; rep<N; ++rep) {
            gettimeofday(&start,0);
            for (int i=0; i<Nqueries; ++i) {
                lqMapOverAllObjectsInLocality(db, cxpos[i], cypos[i], czpos[i], distance, &callbackFunctionLQ, 0);
            }
            gettimeofday(&stop,0);
            // insert-sort the new rate
            FloatType reprate = getRate(start, stop, Nqueries);
            for (int j=0; j<rep; ++j) if (rate[j]<reprate) {
                FloatType tmp = rate[j]; rate[j] = reprate; reprate = tmp;
            }
            rate[rep] = reprate;
        }
        avgrate = 0.f;
        for (int i=0; i<30; ++i) avgrate += rate[i];
        avgrate /= 30.f;
        queryValues << " " << avgrate << flush;
        cout << "." << flush;

#ifdef NEIGHAND_SELECT_METHOD_STAT
        nhCombi.resetStat();
#endif
        for (int rep = 0; rep<N; ++rep) {
            gettimeofday(&start,0);
            for (int i=0; i<Nqueries; ++i) {
                nhCombi.applyToNeighbors(cxpos[i], cypos[i], czpos[i], distance, &callbackFunctionNH, 0);
            }
            gettimeofday(&stop,0);
            // insert-sort the new rate
            FloatType reprate = getRate(start, stop, Nqueries);
            for (int j=0; j<rep; ++j) if (rate[j]<reprate) {
                FloatType tmp = rate[j]; rate[j] = reprate; reprate = tmp;
            }
            rate[rep] = reprate;
        }
        avgrate = 0.f;
        for (int i=0; i<30; ++i) avgrate += rate[i];
        avgrate /= 30.f;
        queryValues << " " << avgrate << flush;
        cout << "." << flush;

#ifdef NEIGHAND_SELECT_METHOD_STAT
cout<< " " << ((float)nhCombi.statSphere / (float)(N*Nqueries))
    << " " << ((float)nhCombi.statParallelepiped / (float)(N*Nqueries))
    << " " << ((float)nhCombi.statNEList / (float)(N*Nqueries)) << flush;
#endif


        queryValues << endl;
        closestValues << endl;
        cout << endl;
    }



    closestValues.close();

    return 0;
}
